import random
import sys
import xxhash
from collections import Counter, OrderedDict

import numpy as np
import pandas as pd
from dsbox.ml.metrics import generalized_jaccard_similarity_score
from dsbox.ml.text.utils import shinglelize
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array, check_is_fitted

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class LSH(BaseEstimator):
    """ Local Sensitive Hashing classifier for texts.

    Comparison of texts can need heavily computing if done with brute force, as string comparison is 
    costly. The problem of finding textually similar documents can be turned into such a set problem 
    by the technique known as shingling. Then, using a technique called minhashing, the classifier
    compresses large sets in such a way that we can still deduce the similarity of the underlying sets
    from their compressed versions, approximating Jaccard similarity coefficient.
    Another important problem that arises when we search for similar items of any kind is that there 
    may be far too many pairs of items to test each pair for their degree of similarity, even if 
    computing the similarity of any one pair can be made very easy. That concern motivates a technique 
    called locality-sensitive hashing, for focusing our search on pairs that are most likely to be similar.


            Parameters
            ----------
            n_hash_functions: int, optional, default 100
                amount of hash functions to use

            shingle_size: int, optional, default 3
                size of shingles

            most_common: int, optional, default 10
                amount of candidates to return during the candidates retrieval phase (hash collisions)

            n_bands: int, optional, default 50
                amount of bands to use for merging hash codes. Must be a multiple of the amount of hash functions

            compare_function: callable, optional, default dsbox.ml.metrics.generalized_jaccard_similarity_score
                metric used to sort found candidates

            Attributes
            ----------
            string_minhashes_ : array
                list of minhashes for each hash function for each shinglelized text

            band_size_ : int
                size of one band

            hash_neighbors_bands_ : list
                buckets of hashes merged into bands

            candidates_ : list of tuples
                full results of predict method with all candidates per text with associated scores

            References
            ----------
            .. [1] Jure Leskovec, Anand Rajaraman, Jeff Ullman. Mining of Massive Datasets, Chapter 3:
            Finding Similar Items, Cambridge University Press, 2011

            Examples
            --------  
            >>> from dsbox.ml.text import LSH
            >>> import random
            >>> random.seed(42) # for making doctests stable

            >>> texts = ['there is a better way', 'demain, avant après-demain, après hier']
            >>> texts_to_compare = ['hier, dès demain']
            >>> clf = LSH()
            >>> clf.fit(texts)
            >>> clf.predict_proba(texts_to_compare)
                                                    0         1
            0  demain, avant après-demain, après hier  0.421053


    """

    def __init__(self, n_hash_functions=100, shingle_size=3,
                 most_common=10, n_bands=50, compare_function=generalized_jaccard_similarity_score):
        try:
            if n_hash_functions % n_bands > 0:
                raise ValueError("Amount of bands must be a multiple of the amount of hash functions")
        except:
            print(str(n_hash_functions) + " divide by " + str(n_bands) + " gives a float value, not int.")

        self._seeds = []
        self._shingle_size = shingle_size
        self.n_bands = n_bands
        self.most_common = most_common
        self.compare_function = compare_function

        self._hashfunctions = []
        self.n_hash_functions = n_hash_functions
        self._init_hash_functions(n_hash_functions)

        self.attr_to_check = ["string_minhashes_",
                              "band_size_",
                              "hash_neighbors_bands_",
                              "X_"]

    def _init_hash_functions(self, n):
        for i in range(0, n):
            self._seeds.append(random.randint(1, 1000000))

    def _get_hash(self, index_function, value):
        x = xxhash.xxh64(value, seed=self._seeds[index_function])
        x = x.intdigest()

        return x

    def _minhash(self, shingles):
        hashes = np.zeros(self.n_hash_functions)

        for i in range(0, self.n_hash_functions):
            min_hash = sys.maxsize
            for j in range(0, len(shingles)):
                val_hash = self._get_hash(i, shingles[j])
                if val_hash < min_hash:
                    min_hash = val_hash
            hashes[i] = min_hash

        return hashes

    def _build_hashes(self, string_vector, shingle_size):
        string_minhashes = np.array(np.zeros(len(string_vector)), dtype=object)
        for i in range(0, len(string_vector)):
            string_minhashes[i] = self._minhash(shinglelize(string_vector[i], shingle_size))

        return string_minhashes

    def _reverse_hash(self, minhashes, n_bands, band_size):

        hash_neighbors_bands = []

        for band in range(0, n_bands):

            hash_dict = {}

            for i in range(0, len(minhashes)):

                hash_code_band = 0
                for h in range(int(band) * int(band_size), (int(band) + 1) * int(band_size)):
                    hash_code_band += minhashes[i][h]

                if hash_code_band not in hash_dict:
                    s = set()
                    s.add(i)
                    hash_dict[hash_code_band] = s
                else:
                    hash_dict[hash_code_band].add(i)

            hash_neighbors_bands.append(hash_dict)

        return hash_neighbors_bands

    def fit(self, X, y=None):
        """ Fit the model by shingling and hashing strings.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        y : array-like, shape (n_samples,)
            No used, only here for compatibility reason

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        check_array(X, ensure_2d=False, dtype='str')

        self.X_ = X

        if self.n_bands is None:
            self.n_bands = self.n_hash_functions

        self.band_size_ = self.n_hash_functions / self.n_bands

        self.string_minhashes_ = self._build_hashes(X, self._shingle_size)

        self.hash_neighbors_bands_ = self._reverse_hash(self.string_minhashes_, self.n_bands, self.band_size_)

    def _get_candidates(self, raw_documents):
        """ For each given string, return a list of the closest strings known by the estimator. 
        Return results are ordered by the amount of "collisions" with the hash keys of the given string each
        known string made.

        Parameters
        ----------
        raw_documents : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        C : array-like, iterable
            Collection of str
        """

        string_minhashes_to_predict = self._build_hashes(raw_documents, self._shingle_size)
        hash_neighbors_bands_to_predict = self._reverse_hash(string_minhashes_to_predict, self.n_bands, self.band_size_)

        documents_candidates = []
        for i in range(0, len(raw_documents)):
            documents_candidates.append(list())

        for band in range(0, len(hash_neighbors_bands_to_predict)):
            candidates = []
            for hash_code_band_to_predict, ids_doc in hash_neighbors_bands_to_predict[band].items():
                if hash_code_band_to_predict in self.hash_neighbors_bands_[band]:
                    candidates = candidates + list(self.hash_neighbors_bands_[band].get(hash_code_band_to_predict))

                for id in ids_doc:
                    documents_candidates[id] = documents_candidates[id] + candidates

        for i in range(0, len(documents_candidates)):
            documents_candidates[i] = Counter(documents_candidates[i]).most_common(self.most_common)
            documents_candidates[i].sort(key=lambda tup: tup[1], reverse=True)
            candidates = list()
            for k in documents_candidates[i]:
                candidates.append(k[0])
            documents_candidates[i] = candidates

        return documents_candidates

    def predict_proba(self, X):
        """Return the closest learned candidate for each text given in input, associated to metric score
        passed in self.compare_function.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        df_candidates : dataframe
            Text candidates with metric score.
        """

        check_is_fitted(self, self.attr_to_check)

        candidates_indexes = self._get_candidates(X)
        candidates = []
        for i in range(0, len(X)):
            candidates.append(dict())

        for i in range(0, len(X)):
            candidate_index = candidates_indexes[i]
            for j in candidate_index:
                candidate = self.X_[j]
                candidates[i][candidate] = self.compare_function(X[i], candidate)

        df_list = []

        for i in range(0, len(X)):
            candidates[i] = list(OrderedDict(sorted(candidates[i].items(), key=lambda t: t[1], reverse=True)).items())
            df_list.append(pd.DataFrame([candidates[i][0]]))

        self.candidates_ = candidates

        df_candidates = pd.concat(df_list, ignore_index=True)

        return df_candidates

    def predict(self, X):
        """Return the closest learned candidate for each text given in input.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        df_candidates : array-like, shape (n_samples,)
            Text candidates.
        """

        df_candidates = self.predict_proba(X)
        return df_candidates[0]



import unittest
from numpy.testing import assert_array_equal, assert_array_almost_equal, assert_raises

import numpy as np

from dsbox.ml.markov.seqlearn import MultinomialHMM


class TestSeqHMM(unittest.TestCase):
    text = [w.split() for w in ["this DT",
                                "is V",
                                "a DT",
                                "test N",
                                "for IN",
                                "a DT",
                                "hidden Adj",
                                "Markov N",
                                "model N"]]
    words, y = zip(*text)
    lengths = [len(text)]

    vocab, identities = np.unique(words, return_inverse=True)
    X = (identities.reshape(-1, 1) == np.arange(len(vocab))).astype(int)

    def test_hmm(self):
        n_features = self.X.shape[1]

        clf = MultinomialHMM()
        clf.fit(self.X, self.y, self.lengths)
        assert_array_equal(clf.classes_, ["Adj", "DT", "IN", "N", "V"])
        assert_array_equal(clf.predict(self.X), self.y)

        clf.set_params(decode="bestfirst")
        assert_array_equal(clf.predict(self.X), self.y)

        n_classes = len(clf.classes_)
        assert_array_almost_equal(np.ones(n_features),
                                  np.exp(clf.coef_).sum(axis=0))
        assert_array_almost_equal(np.ones(n_classes),
                                  np.exp(clf.intercept_trans_).sum(axis=0))
        assert_array_almost_equal(1., np.exp(clf.intercept_final_).sum())
        assert_array_almost_equal(1., np.exp(clf.intercept_init_).sum())

    def test_hmm_validation(self):
        assert_raises(ValueError, MultinomialHMM(alpha=0).fit, self.X, self.y, self.lengths)
        assert_raises(ValueError, MultinomialHMM(alpha=-1).fit, self.X, self.y, self.lengths)



from airflow.providers.cncf.kubernetes.operators.kubernetes_pod import KubernetesPodOperator
from airflow.providers.google.cloud.operators.kubernetes_engine import GKEStartPodOperator


class KubeDataOperator(KubernetesPodOperator):
    """

    Parameters
    ----------
    operation: str
        operation name defined in the data catalog operations used in the project

    kube_conf: dict, default=None
        Kubernetes cluster configuration. Must contain at least 'arguments' and 'cmds' keys

    kwargs: dict
        Other Parameters passed to KubernetesPodOperator constructor
    """

    ui_color = '#defdff'

    def __init__(self, operation, kube_conf=None, **kwargs):
        self.operation = operation
        kube_conf['arguments'] += [operation]
        kube_conf['task_id'] = operation
        kube_conf.update(kwargs)

        name = '{}-{}-pod'.format(kube_conf['dag'].dag_id, operation).replace('_', '-').lower()[0:63]
        kube_conf['name'] = name

        super().__init__(**kube_conf)


class GKEDataOperator(GKEStartPodOperator):
    """

    Parameters
    ----------
    operation: str
        operation name defined in the data catalog operations used in the project

    kube_conf: dict, default=None
        Kubernetes cluster configuration. Must contain at least 'arguments' and 'cmds' keys

    kwargs: dict
        Other Parameters passed to GKEStartPodOperator constructor
    """

    ui_color = '#defdff'

    def __init__(self, operation, kube_conf=None, **kwargs):
        self.operation = operation
        kube_conf['arguments'] += [operation]
        kube_conf['task_id'] = operation
        kube_conf.update(kwargs)

        name = '{}-{}-pod'.format(kube_conf['dag'].dag_id, operation).replace('_', '-').lower()[0:63]

        kube_conf['name'] = name

        super().__init__(**kube_conf)



import matplotlib.pyplot as plt
import numpy as np

__author__ = "Samuel Rochette"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def plot_eval_prediction_interval(target_eval, prediction, lower_error_list, upper_error_list, figsize=(10, 7)):
    """
    Display the sorted predictions of the evaluation set and their prediction intervals versus the ground truth.

    Parameters
    ----------
    target_eval : array-like
    prediction : array-like, prediction of the evaluation set
    lower_error_list : array-like
    upper_error_list : array-like
    figsize : tuple, optional (default=(10,7)), set the figure size

    Returns
    -------
    Prediction interval figure

    References
    ----------
    [1]  Quantile Regression Forests, Nicolai Meinshausen. 2006
    [2]  https://blog.datadive.net/prediction-intervals-for-random-forests/

    Examples
    --------
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from dsbox.ml.explain.prediction_interval import compute_prediction_interval
    >>> from dsbox.ml.visualization.prediction_interval_visualization import plot_eval_prediction_interval

    >>> boston = load_boston()
    >>> data_train, data_eval, target_train, target_eval = train_test_split(boston['data'], boston['target'], test_size=0.3, random_state=42)
    
    >>> rf = RandomForestRegressor(n_estimators=100, n_jobs=-1)
    >>> _ = rf.fit(data_train, target_train)
    
    >>> lower_error_list, upper_error_list = compute_prediction_interval(rf, data_eval, percentile=90)
    >>> prediction = rf.predict(data_eval)
    >>> plot_eval_prediction_interval(target_eval, prediction, lower_error_list, upper_error_list)
    """

    x = range(len(prediction))
    permutation_sorted_ground_truth = target_eval.argsort()
    y = target_eval[permutation_sorted_ground_truth]
    sorted_predictions = prediction[permutation_sorted_ground_truth]
    upper_error = np.array(upper_error_list)[permutation_sorted_ground_truth] - sorted_predictions
    lower_error = sorted_predictions - np.array(lower_error_list)[permutation_sorted_ground_truth]

    plt.figure(figsize=figsize)
    plt.plot(x, y, color='c', marker='o', markersize=6, label='True values', linestyle='None')
    plt.errorbar(x, sorted_predictions, yerr=[lower_error, upper_error], fmt='bo', label='Predicted values', capsize=3)
    plt.ylabel('Response value')
    plt.xlabel('Sample no.')
    plt.legend(loc='upper left')
    plt.show()


def plot_prediction_interval(prediction, lower_error_list, upper_error_list, figsize=(10, 7)):
    """
    Display the sorted predictions their prediction intervals.

    Parameters
    ----------
    prediction : array-like, prediction of the evaluation set
    lower_error_list : array-like
    upper_error_list : array-like
    figsize : tuple, optional (default=(10,7)), set the figure size

    Returns
    -------
    Prediction interval figure

    Examples
    --------
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from dsbox.ml.explain.prediction_interval import compute_prediction_interval
    >>> from dsbox.ml.visualization.prediction_interval_visualization import plot_prediction_interval

    >>> boston = load_boston()
    >>> data_train, data_eval, target_train, target_eval = train_test_split(boston['data'], boston['target'], test_size=0.3, random_state=42)
    >>> rf = RandomForestRegressor(n_estimators=100, n_jobs=-1)
    >>> _ = rf.fit(data_train, target_train)

    >>> lower_error_list, upper_error_list = compute_prediction_interval(rf, data_eval, percentile=90)
    >>> prediction = rf.predict(data_eval)

    >>> plot_prediction_interval(prediction, lower_error_list, upper_error_list)
    """

    x = range(len(prediction))
    permutation_sorted_prediction = prediction.argsort()
    sorted_predictions = prediction[permutation_sorted_prediction]
    upper_error = np.array(upper_error_list)[permutation_sorted_prediction] - sorted_predictions
    lower_error = sorted_predictions - np.array(lower_error_list)[permutation_sorted_prediction]

    plt.figure(figsize=figsize)
    plt.errorbar(x, sorted_predictions, yerr=[lower_error, upper_error], fmt='bo', label='Predicted values', capsize=3)
    plt.ylabel('Response value')
    plt.xlabel('Sample no.')
    plt.legend(loc='upper left')
    plt.show()



import os
import pyarrow.plasma as plasma

test_user = os.environ.get(['TEST_DSBOX_PG_USER'])
test_password = os.environ.get(['TEST_DSBOX_PG_PASSWD'])
test_hostname = os.environ.get(['TEST_DSBOX_PG_HOST'])
test_port = os.environ.get(['TEST_DSBOX_PG_PORT'])
test_dbname = os.environ.get(['TEST_DSBOX_PG_DBNAME'])

socket_name = os.environ.get(['TEST_DSBOX_PLASMA_SOCKET_NAME'])

id = bytes([0x01] * 20)
object_id = plasma.ObjectID(id)


import unittest
from datetime import datetime
import pandas as pd

from airflow import DAG

from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputFileUnit, DataGlobalInputUnit, \
    DataGlobalOutputFileUnit


def drop_na_dataframe(dataframe, columns):
    dataframe = dataframe.dropna(subset=columns)
    return dataframe


def drop_na_vaex_dataframe(dataframe, columns):
    dataframe = dataframe.dropna(column_names=columns)
    return dataframe


class TestDataOperator(unittest.TestCase):
    def test_execute_data_operator_csv_read_and_write(self):
        # given
        dag = DAG(dag_id='test', start_date=datetime.now())
        input_csv_unit = DataInputFileUnit('data/X.csv', sep=';')
        output_csv_unit = DataOutputFileUnit('data/X_parsed.csv', index=False)

        task = DataOperator(operation_function=drop_na_dataframe,
                            params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                            input_unit=input_csv_unit,
                            output_unit=output_csv_unit,
                            dag=dag, task_id='data_operator_csv')

        # when
        task.execute(None)

        # then
        df_transformed = pd.read_csv('data/X_parsed.csv')
        self.assertEqual((10245, 27), df_transformed.shape)

    def test_execute_data_operator_csv_read_and_parquet_write(self):
        # given
        dag = DAG(dag_id='test', start_date=datetime.now())
        input_csv_unit = DataInputFileUnit('data/X.csv', sep=';')
        output_parquet_unit = DataOutputFileUnit('data/X_parsed.parquet', pandas_write_function_name='to_parquet')

        task = DataOperator(operation_function=drop_na_dataframe,
                            params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                            input_unit=input_csv_unit,
                            output_unit=output_parquet_unit,
                            dag=dag, task_id='data_operator_csv_to_parquet')

        # when
        task.execute(None)

        # then
        df_transformed = pd.read_parquet('data/X_parsed.parquet', engine='pyarrow')
        self.assertEqual((10245, 27), df_transformed.shape)

    def test_execute_data_operator_csv_read_and_parquet_write_using_vaex_api_backend(self):
        # given
        dag = DAG(dag_id='test', start_date=datetime.now())
        input_csv_unit = DataGlobalInputUnit('data/X.csv', api_module='vaex', sep=';')
        output_parquet_unit = DataGlobalOutputFileUnit('data/X_parsed.parquet', write_function_name='export_parquet')

        task = DataOperator(operation_function=drop_na_vaex_dataframe,
                            params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                            input_unit=input_csv_unit,
                            output_unit=output_parquet_unit,
                            dag=dag, task_id='data_operator_csv_to_parquet')

        # when
        task.execute(None)

        # then
        df_transformed = pd.read_parquet('data/X_parsed.parquet', engine='pyarrow')
        self.assertEqual((10245, 27), df_transformed.shape)

    def test_execute_data_operator_csv_read_and_parquet_write_using_dask_api_backend(self):
        # given
        dag = DAG(dag_id='test', start_date=datetime.now())
        input_csv_unit = DataGlobalInputUnit('data/X.csv', api_module='dask.dataframe', sep=';')
        output_parquet_unit = DataGlobalOutputFileUnit('data/X_parsed', write_function_name='to_parquet',
                                                       write_index=False)

        task = DataOperator(operation_function=drop_na_dataframe,
                            params={'columns': ['ANNEEREALISATIONDIAGNOSTIC']},
                            input_unit=input_csv_unit,
                            output_unit=output_parquet_unit,
                            dag=dag, task_id='data_operator_csv_to_parquet')

        # when
        task.execute(None)

        # then
        df_transformed = pd.read_parquet('data/X_parsed', engine='pyarrow')
        self.assertEqual((10245, 27), df_transformed.shape)


if __name__ == '__main__':
    unittest.main()



import pandas as pd

from dsbox.ml.feature_engineering import TagEncoder
from dsbox.ml.feature_engineering.timeseries import RollingWindower, Shifter
from dsbox.utils import pandas_downcast_numeric


def concat_train_test(dataframe_list):
    shop_data = dataframe_list[0]
    shop_data_to_predict = dataframe_list[1]
    shop_data['to_predict'] = False
    shop_data_to_predict['to_predict'] = True
    shop_data = pd.concat([shop_data, shop_data_to_predict], sort=False)
    shop_data = shop_data.reset_index(drop=True)
    return shop_data


def resample_fillna(dataframe):
    shop_data = dataframe
    shop_data['Datetime'] = pd.to_datetime(shop_data['Date'], format='%Y-%m-%d')
    shop_data = shop_data.set_index('Datetime')
    print("Resampling...")
    shop_data = shop_data.groupby('Store').resample('D').asfreq()
    shop_data = shop_data.drop('Store', axis=1)
    shop_data = shop_data.reset_index(0)

    print("Fill missing values...")
    shop_data['MissingValues'] = shop_data['Sales'].isnull()

    shop_data['DayOfWeek'] = shop_data.index.map(lambda x: x.dayofweek)
    shop_data['Month'] = shop_data.index.map(lambda x: x.month)
    sales_per_store_month_dayofweek = shop_data.groupby(['Store', 'Month', 'DayOfWeek'])['Sales'].median()
    customers_per_store_month_dayofweek = shop_data.groupby(['Store', 'Month', 'DayOfWeek'])['Customers'].median()
    dict_sales_per_store_month_dayofweek = sales_per_store_month_dayofweek.to_dict()
    dict_customers_per_store_month_dayofweek = customers_per_store_month_dayofweek.to_dict()

    shop_data['T_StoreMonthDow'] = list(zip(shop_data['Store'], shop_data['Month'], shop_data['DayOfWeek']))
    shop_data['SalesNew'] = shop_data['T_StoreMonthDow'].map(lambda x: dict_sales_per_store_month_dayofweek[x])
    shop_data['CustomersNew'] = shop_data['T_StoreMonthDow'].map(lambda x: dict_customers_per_store_month_dayofweek[x])
    shop_data['SalesNew'] = shop_data['SalesNew'] * shop_data['MissingValues']
    shop_data['CustomersNew'] = shop_data['CustomersNew'] * shop_data['MissingValues']
    shop_data['Sales'] = shop_data['Sales'].fillna(0) + shop_data['SalesNew']
    shop_data['Customers'] = shop_data['Customers'].fillna(0) + shop_data['CustomersNew']
    shop_data = shop_data.drop(['SalesNew', 'CustomersNew', 'T_StoreMonthDow'], axis=1)

    shop_data = shop_data.reset_index(drop=False)

    pandas_downcast_numeric(shop_data)

    print("Remaining missing values:\n" + str(shop_data[['Sales', 'Customers']].isnull().sum()))

    return shop_data


def date_to_dateint(date_val):
    return date_val.year * 10000 + date_val.month * 100 + date_val.day


def create_simple_features(dataframe):
    X = dataframe

    X['Date'] = X['Datetime'].apply(lambda x: date_to_dateint(x))
    X['Week'] = X['Datetime'].apply(lambda x: x.week)
    X['Year'] = X['Datetime'].apply(lambda x: x.year)

    X['SalePerCustomer'] = X['Sales'] / X['Customers']
    X['SalePerCustomer'] = X['SalePerCustomer'].fillna(0)

    col_list = ['StateHoliday']
    for col in col_list:
        te = TagEncoder()
        X[col] = te.fit_transform(X[col].fillna('nan'))

    pandas_downcast_numeric(X)

    return X


def create_timeseries_rolling_features(X_ts, operation='mean', features=['Sales'], shift_ranges=[42],
                                       rolling_windows=[7]):
    print(str(operation))
    X_ts = X_ts.sort_values(['Store', 'Date'])
    roller = RollingWindower(operation=operation, windows=rolling_windows, min_periods=1)
    X_roll = roller.transform(X_ts.groupby('Store')[features]).reset_index(0, drop=False)
    roll_shifter = Shifter(shifts=shift_ranges, prefix='d-')
    X_roll_shifted = roll_shifter.transform(X_roll.groupby('Store'))
    X_ts = X_ts.join(X_roll_shifted)
    X_ts = X_ts.drop(features, axis=1)

    return X_ts


def create_timeseries_shift_features(X_ts, features=['Sales'], shift_ranges=[42]):
    print("shift features")
    X_ts = X_ts.sort_values(['Store', 'Date'])
    shifter = Shifter(shifts=shift_ranges, prefix='d-')
    X_shifted = shifter.transform(X_ts.groupby('Store')[features])
    X_ts = X_ts.join(X_shifted)
    X_ts = X_ts.drop(features, axis=1)

    return X_ts


def create_timeseries_diff_shift_features(df_shift, features=['Sales'], shift_ranges=[42, 47], prefix='diff_'):
    for feat in features:
        for i in range(0, len(shift_ranges) - 1):
            df_shift[prefix + 'd-' + str(shift_ranges[i + 1]) + '_' + str(shift_ranges[i]) + '_' + feat] = \
                df_shift['d-' + str(shift_ranges[i + 1]) + '_' + feat] - df_shift[
                    'd-' + str(shift_ranges[i]) + '_' + feat]

    return df_shift


def merge_time_features(dataframe_list, features=['Sales'], merge_cols=['Store', 'Date']):
    for i in range(0, len(dataframe_list)):
        dataframe_list[i] = dataframe_list[i].sort_values(merge_cols).reset_index(drop=True)

    print("Merge...")
    passed_features = dataframe_list[0]
    for i in range(1, len(dataframe_list)):
        print('df ' + str(i))
        for col in dataframe_list[i].columns:
            if col not in merge_cols:
                print(col)
                passed_features[col] = dataframe_list[i][col]

    print(passed_features.columns)

    passed_features = passed_features.fillna(method='backfill')
    passed_features = passed_features.fillna(method='ffill')

    return passed_features


def extra_features(dataframe):
    extra_data = dataframe
    cols_to_fill = ['CompetitionDistance', 'CompetitionOpenSinceMonth', 'CompetitionOpenSinceYear', 'Promo2SinceWeek',
                    'Promo2SinceYear']
    for col in cols_to_fill:
        extra_data[col] = extra_data[col].fillna(extra_data[col].median())

    extra_data = extra_data.fillna(0)
    for col in ['StoreType', 'Assortment', 'PromoInterval']:
        te = TagEncoder()
        extra_data[col] = te.fit_transform(extra_data[col].fillna('nan'))

    for col in cols_to_fill:
        extra_data[col] = extra_data[col].astype('int')

    pandas_downcast_numeric(extra_data)

    return extra_data


def merge_final_features(dataframe_list):
    X = dataframe_list[0].merge(dataframe_list[1], on='Store')

    X['CompetitionOpen'] = 12 * (X['Year'] - X['CompetitionOpenSinceYear']) + \
                           (X['Month'] - X['CompetitionOpenSinceMonth'])
    X['PromoOpen'] = 12 * (X['Year'] - X['Promo2SinceYear']) + \
                     (X['Week'] - X['Promo2SinceWeek']) / 4.0

    return X


"""Private utilities."""

import numpy as np
from sklearn.utils.validation import NotFittedError


# Copied from scikit-learn 0.19.
def _validate_covars(covars, covariance_type, n_components):
    """Do basic checks on matrix covariance sizes and values."""
    from scipy import linalg
    if covariance_type == 'spherical':
        if len(covars) != n_components:
            raise ValueError("'spherical' covars have length n_components")
        elif np.any(covars <= 0):
            raise ValueError("'spherical' covars must be positive")
    elif covariance_type == 'tied':
        if covars.shape[0] != covars.shape[1]:
            raise ValueError("'tied' covars must have shape (n_dim, n_dim)")
        elif (not np.allclose(covars, covars.T)
              or np.any(linalg.eigvalsh(covars) <= 0)):
            raise ValueError("'tied' covars must be symmetric, "
                             "positive-definite")
    elif covariance_type == 'diag':
        if len(covars.shape) != 2:
            raise ValueError("'diag' covars must have shape "
                             "(n_components, n_dim)")
        elif np.any(covars <= 0):
            raise ValueError("'diag' covars must be positive")
    elif covariance_type == 'full':
        if len(covars.shape) != 3:
            raise ValueError("'full' covars must have shape "
                             "(n_components, n_dim, n_dim)")
        elif covars.shape[1] != covars.shape[2]:
            raise ValueError("'full' covars must have shape "
                             "(n_components, n_dim, n_dim)")
        for n, cv in enumerate(covars):
            if (not np.allclose(cv, cv.T)
                    or np.any(linalg.eigvalsh(cv) <= 0)):
                raise ValueError("component %d of 'full' covars must be "
                                 "symmetric, positive-definite" % n)
    else:
        raise ValueError("covariance_type must be one of " +
                         "'spherical', 'tied', 'diag', 'full'")


# Copied from scikit-learn 0.19.
def distribute_covar_matrix_to_match_covariance_type(
        tied_cv, covariance_type, n_components):
    """Create all the covariance matrices from a given template."""
    if covariance_type == 'spherical':
        cv = np.tile(tied_cv.mean() * np.ones(tied_cv.shape[1]),
                     (n_components, 1))
    elif covariance_type == 'tied':
        cv = tied_cv
    elif covariance_type == 'diag':
        cv = np.tile(np.diag(tied_cv), (n_components, 1))
    elif covariance_type == 'full':
        cv = np.tile(tied_cv, (n_components, 1, 1))
    else:
        raise ValueError("covariance_type must be one of " +
                         "'spherical', 'tied', 'diag', 'full'")
    return cv


# Adapted from scikit-learn 0.21.
def check_is_fitted(estimator, attribute):
    if not hasattr(estimator, attribute):
        raise NotFittedError(
            "This %s instance is not fitted yet. Call 'fit' with "
            "appropriate arguments before using this method."
            % type(estimator).__name__)


from dsbox.ml.neural_networks import KerasFactory
from tensorflow.keras.layers import Dense, Activation, Embedding, LSTM
from tensorflow.keras.layers import Dropout, Conv1D, MaxPooling1D
from tensorflow.keras.models import Sequential

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class LSTMFactory(KerasFactory):
    def create_model(self, dictionary_size, num_labels=2):
        """
        Build a keras model with several layers:
        
        - Embedding
        - LSTM
        - Dense
        - Activation
        
        Mostly used for text classification.
        
        Parameters
        ----------
        dictionary_size: int
            Size of the vocabulary
        num_labels: int, optional (default=2)
            Number of classes.

        Returns
        -------
        Keras model

        """
        model = Sequential()
        model.add(Embedding(dictionary_size, 128))
        model.add(LSTM(32, dropout=0.1, recurrent_dropout=0.1))
        model.add(Dense(num_labels))
        model.add(Activation('softmax'))
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        return model


class CNN_LSTMFactory(KerasFactory):
    """
        Build a keras model with several layers:

        - Embedding
        - Dropout
        - Conv1D
        - MaxPooling1D
        - LSTM
        - Dense
        - Activation

        Mostly used for text classification.

        Parameters
        ----------
        dictionary_size: int
            Size of the vocabulary
        num_labels: int, optional (default=2)
            Number of classes.

        Returns
        -------
        Keras model

        """

    def create_model(self, dictionary_size, num_labels=2):
        model = Sequential()
        model.add(Embedding(dictionary_size, 128))
        model.add(Dropout(0.1))
        model.add(Conv1D(64,
                         5,
                         padding='valid',
                         activation='relu',
                         strides=1))
        model.add(MaxPooling1D(pool_size=4))
        model.add(LSTM(64, dropout=0.1, recurrent_dropout=0.1))
        model.add(Dense(num_labels))
        model.add(Activation('softmax'))
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        return model



"""Hidden Markov models (HMMs) with supervised training."""

# Copyright 2013 Lars Buitinck

import numpy as np
from scipy.special import logsumexp

from dsbox.ml.markov.seqlearn.base import BaseSequenceClassifier
from dsbox.ml.markov.seqlearn._utils import atleast2d_or_csr, count_trans, safe_sparse_dot

__credits__ = "Lars Buitinck"

class MultinomialHMM(BaseSequenceClassifier):
    """First-order hidden Markov model with multinomial event model.

    Parameters
    ----------
    decode : string, optional
        Decoding algorithm, either "bestfirst" or "viterbi" (default).
        Best-first decoding is also called posterior decoding in the HMM
        literature.

    alpha : float
        Lidstone (additive) smoothing parameter.
        
    Examples
    --------
    
    >>> import pandas as pd
    >>> from dsbox.ml.markov.seqlearn import MultinomialHMM
    >>> clf = MultinomialHMM()
    >>> X_train = pd.DataFrame({'a': [0, 1, 2], 'b':[0, 1, 2], 'c': [0, 1, 2]})
    >>> y_train = np.array([0, 1, 2])
    >>> X_test = pd.DataFrame({'a': [1, 1, 1], 'b':[1, 1, 1], 'c': [1, 1, 1]})
    >>> _ = clf.fit(X_train, y_train, [len(y_train)])
    >>> clf.predict(X_test)
    array([2, 1, 2])
    """

    def __init__(self, decode="viterbi", alpha=.01):
        self.alpha = alpha
        self.decode = decode

    def fit(self, X, y, lengths):
        """Fit HMM model to data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Feature matrix of individual samples.

        y : array-like, shape (n_samples,)
            Target labels.

        lengths : array-like of integers, shape (n_sequences,)
            Lengths of the individual sequences in X, y. The sum of these
            should be n_samples.

        Notes
        -----
        Make sure the training set (X) is one-hot encoded; if more than one
        feature in X is on, the emission probabilities will be multiplied.

        Returns
        -------
        self : MultinomialHMM
        """

        alpha = self.alpha
        if alpha <= 0:
            raise ValueError("alpha should be >0, got {0!r}".format(alpha))

        X = atleast2d_or_csr(X)
        classes, y = np.unique(y, return_inverse=True)
        lengths = np.asarray(lengths)
        Y = y.reshape(-1, 1) == np.arange(len(classes))

        end = np.cumsum(lengths)
        start = end - lengths

        init_prob = np.log(Y[start].sum(axis=0) + alpha)
        init_prob -= logsumexp(init_prob)
        final_prob = np.log(Y[start].sum(axis=0) + alpha)
        final_prob -= logsumexp(final_prob)

        feature_prob = np.log(safe_sparse_dot(Y.T, X) + alpha)
        feature_prob -= logsumexp(feature_prob, axis=0)

        trans_prob = np.log(count_trans(y, len(classes)) + alpha)
        trans_prob -= logsumexp(trans_prob, axis=0)

        self.coef_ = feature_prob
        self.intercept_init_ = init_prob
        self.intercept_final_ = final_prob
        self.intercept_trans_ = trans_prob

        self.classes_ = classes

        return self



import numpy as np
import pandas as pd
from dsbox.ml.markov.histrogramdata import HistogramData
from sklearn.base import BaseEstimator
from sklearn.utils.validation import column_or_1d, check_is_fitted

__author__ = "Vincent Levorato, Michel Lutz"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class MarkovSequenceMiner(BaseEstimator):
    """ Simple Markov Chain model.

    In principle, when we observe a sequence of chance experiments, all of the past outcomes could
    influence our predictions for the next experiment : the outcome of a given experiment can affect 
    the outcome of the next experiment. This type of process is called a Markov chain.

    The process starts in one of the states of a sequence and moves successively from one state to 
    another. If the chain is currently in state 'A, then it moves to state 'B' at the next step with
    a probability denoted by p_AB. These probabilities are stored in a transition matrix.

    The MarkovSequenceMiner builds the transition matrix with the fit() method. It's possible to test
    another sequence to get the probabilities of having a new state knowing only the last one or to get 
    the probabilities of having a new state knowing all the previous ones. 

    Parameters
    ----------
    init_label : str, optional (default=None)
        Gives explicitly the first state of the sequence. If not set, the first state will be the first
        element seen during the fit process.

    warm_start : bool, optional (default=False)
        When set to ``True``, remember the previous states seen during the previous predict process
        when calling the predict process.

    equiproba : bool, optional (default=False)
        When set to ``True``, ignore the init_label given value, and gives to the init vector the same
        probabilites weight (i.e. if there are 3 possible states, it gives 1/3 chances to have a state
        for the starting configuration).

    Attributes
    ----------
    labels_ : array of all the states
        Corresponds to states seen during the fit process.

    inv_dict_labels_ : dict
        For a given state, returns its position in the labels_ array and transit_matrix_. 

    transit_matrix_ : matrix
        Corresponds to the Markov transition matrix learned during the fit process.

    Examples
    -------- 
    >>> from dsbox.ml.markov import MarkovSequenceMiner

    >>> x_train = ['a', 'a', 'b', 'c', 'd', 'a']
    >>> clf = MarkovSequenceMiner()
    >>> clf.fit(x_train)
    MarkovSequenceMiner(equiproba=False, init_label='a', warm_start=False)

    >>> x_test = ['a', 'b', 'c', 'd', 'e', 'a']
    >>> chained_p, transition_p = clf.predict_proba(x_test)
    >>> print(chained_p)
    [1.  0.5 0.5 0.5 0.  0. ]
    >>> print(transition_p)
    [1.  0.5 1.  1.  0.  0. ]

    """

    def __init__(self, init_label=None, warm_start=False, equiproba=False):
        self.init_label = init_label
        self.equiproba = equiproba
        self.warm_start = warm_start
        self.init_vector = None
        self.cur_vector = None

        self.attr_to_check = ["labels_",
                              "inv_dict_labels_",
                              "transit_matrix_"
                              ]

    def fit(self, X, y=None):
        """
        Fit the model by estimating probabilities between two states.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str
        y : array-like, shape (n_samples,)
            No used, only here for compatibility reason

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        x = column_or_1d(X)

        if self.init_label is None:
            self.init_label = x[0]

        # get all different labels of the sequence
        # keep their index in an reverse dict and build the transition matrix
        self.labels_ = np.unique(x)
        label_amount = len(self.labels_)
        self.inv_dict_labels_ = {self.labels_[i]: i for i in range(0, label_amount)}
        self.transit_matrix_ = np.zeros([label_amount, label_amount])

        # initialize vector with the first event
        self.init_vector = np.zeros(label_amount)
        self.init_vector[self.inv_dict_labels_[self.init_label]] = 1

        if self.equiproba:
            self.init_vector = np.ones(label_amount) * (1.0 / label_amount)

        for i in range(1, len(x)):
            previous_event_index = self.inv_dict_labels_[x[i - 1]]
            current_event_index = self.inv_dict_labels_[x[i]]
            self.transit_matrix_[previous_event_index, current_event_index] += 1

        for i in range(0, self.transit_matrix_.shape[0]):
            self.transit_matrix_[i, :] = self.transit_matrix_[i, :] / self.transit_matrix_[i, :].sum()

        return self

    def predict_proba(self, X):
        """
        Return two arrays of probabilities for each state transition in sequence X.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        chained_p_ : array-like
            Chained probabilities
        transition_p_ : array-like
            Transition probabilities
        """
        check_is_fitted(self, self.attr_to_check)
        x = column_or_1d(X)

        # initialize the current vector that evolves with each new event
        if not self.warm_start or self.cur_vector is None:
            self.compute_probas_ = np.zeros([len(x), len(self.labels_)])
            self.compute_probas_[0, :] = self.init_vector
            self.cur_vector = np.zeros(len(self.labels_))
            self.cur_vector[self.inv_dict_labels_[x[0]]] = 1
            self.cur_vector = self.cur_vector.reshape(1, len(self.cur_vector))
            self.chained_p_ = np.ones(len(x))
            self.transition_p_ = np.ones(len(x))

        for i in range(1, len(x)):
            self.cur_vector = np.dot(self.cur_vector, self.transit_matrix_)
            self.compute_probas_[i, :] = self.cur_vector[0, :]

            if x[i - 1] not in self.inv_dict_labels_ or x[i] not in self.inv_dict_labels_:
                self.transition_p_[i] = 0.0
                self.chained_p_[i] = 0.0
            else:
                previous_event_index = self.inv_dict_labels_[x[i - 1]]
                current_event_index = self.inv_dict_labels_[x[i]]
                self.transition_p_[i] = self.transit_matrix_[previous_event_index, current_event_index]
                self.chained_p_[i] = self.cur_vector[0, current_event_index]

        return self.chained_p_, self.transition_p_

    def predict(self, X):
        """
        For each transition in sequence X, returns a boolean value, which is True if a transition was the most
         probable transition to occur, and False if not.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        bool_vector : array-like
            Collection of boolean values

        """
        self.predict_proba(X)
        x = column_or_1d(X)
        bool_vector = np.zeros(len(x), dtype=bool)

        for i in range(0, len(x)):
            proba_vector = self.compute_probas_[i]
            max_proba = proba_vector.max()
            i_max_labels = set()
            while max_proba == proba_vector.max():
                i_max = proba_vector.argmax()
                i_max_labels.add(self.labels_[i_max])
                proba_vector[i_max] = -1

            if x[i] in i_max_labels:
                bool_vector[i] = True

        return bool_vector

    def predict_next(self, X):
        """
        For a given sequence X, returns the most probable event(s) to occur.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        predicted_next_event : array-like
            Most probable next event to occur

        """
        self.predict_proba(X)
        proba_vector = self.cur_vector
        proba_vector = np.dot(proba_vector, self.transit_matrix_)
        max_proba = proba_vector.max()
        predicted_next_event = []

        while max_proba == proba_vector.max():
            i_max = proba_vector.argmax()
            predicted_next_event.append(self.labels_[i_max])
            proba_vector[0, i_max] = -1

        return np.array(predicted_next_event)


class TimeDurationSequenceMiner(BaseEstimator):
    """ Sequence time distribution miner.
    
    The purpose of this class is to associate Symbolic data analysis (SDA) to state transitions in a sequence. In 
    this particular case, the time duration distribution is computed and kept for each state transition. It  
    gives information about the normal (or not) behavior on transition, using the time duration distribution point
    of view.
    
    Parameters
    ----------
    bins : int, optional (default=3)
        Set the number of bins to use to estimate the "discrete" distribution using an histogram
        
    
    Attributes
    ----------
    predict_threshold : float, (default=0.5)
        Used by the predict method, to decide if the probablity outputs a True value (strictly above the threshold)
        or a False value (equal or below the threshold)
        
    labels_ : array of all the states
        Corresponds to states seen during the fit process.
    
    inv_dict_labels_ : dict
        For a given state, returns its position in the labels_ array and transit_matrix_.
        
    timeduration_matrix_ : matrix
        Corresponds to the time durations between states (each cell contains a list of values).
        
    sda_matrix_ : matrix
        Corresponds to the time duration distribution between states (each celle contains an HistogramData object).
    
    References
    ----------
    .. [1] Lynne Billard, Edwin Diday. Symbolic Data Analysis: Conceptual Statistics and Data Mining. 
    John Wiley & Sons. ISBN 978-0-470-09017-6, 2012
    
    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.markov import TimeDurationSequenceMiner
    
    >>> df = pd.DataFrame({'event': ['a', 'a', 'b', 'b', 'a', 'a', 'a', 'b'], \
    'timestamp': [0, 12, 14, 29, 40, 42, 44, 48]})
                           
    >>> tds = TimeDurationSequenceMiner()
    >>> tds.fit(df[['event', 'timestamp']].values)
    TimeDurationSequenceMiner(bins=3)
    
    >>> df_test = pd.DataFrame({'event': ['a', 'b', 'b', 'a', 'b', 'a'], \
                                'timestamp': [0, 2, 20, 24, 27, 38]})
                                
    >>> y_test = tds.predict_proba(df_test[['event', 'timestamp']].values)
    >>> print(y_test)
    [1.  0.5 0.  0.  0.  1. ]
    
    
    """

    def __init__(self, bins=3):
        self.bins = bins
        self.predict_threshold = 0.5

        self.attr_to_check = ["labels_",
                              "inv_dict_labels_",
                              "timeduration_matrix_",
                              "sda_matrix_"
                              ]

    def fit(self, X, y=None, log_base=False):
        """
        Fit the model by estimating time duration distribution between states. 

        Parameters
        ----------
        X : array-like, shape = (n_samples, 2)
            First column corresponds to states, second column to timestamps
        y : array-like, shape (n_samples,)
            No used, only here for compatibility reason

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        if X.shape[1] != 2:
            raise ValueError("Shape must be exactly (n,2) but is " + X.shape)

        x = X[:, 0]
        timestamp = X[:, 1]

        x = column_or_1d(x)
        timestamp = column_or_1d(timestamp)

        self.labels_ = np.unique(x)
        label_amount = len(self.labels_)
        self.inv_dict_labels_ = {self.labels_[i]: i for i in range(0, label_amount)}
        self.sda_matrix_ = np.empty([label_amount, label_amount], dtype=object)
        self.timeduration_matrix_ = np.empty([label_amount, label_amount], dtype=object)
        for i in range(0, label_amount):
            for j in range(0, label_amount):
                self.timeduration_matrix_[i, j] = list()

        for i in range(1, len(x)):
            time_duration = timestamp[i] - timestamp[i - 1]
            previous_event_index = self.inv_dict_labels_[x[i - 1]]
            current_event_index = self.inv_dict_labels_[x[i]]
            self.timeduration_matrix_[previous_event_index, current_event_index].append(time_duration)

        for i in range(0, label_amount):
            for j in range(0, label_amount):
                if self.timeduration_matrix_[i, j] is not None:
                    if log_base:
                        hist = np.histogram(np.log(self.timeduration_matrix_[i, j]), bins=self.bins)
                    else:
                        hist = np.histogram(self.timeduration_matrix_[i, j], bins=self.bins)

                    hda = HistogramData(hist[1], hist[0])
                    self.sda_matrix_[i, j] = hda

        return self

    def predict_proba(self, X):
        """
        Return an array corresponding to probabilities for having these time durations
        between states of X.

        Parameters
        ----------
        X : array-like, shape = (n_samples, 2)
            First column corresponds to states, second column to timestamps

        Returns
        -------
        time_p_: array-like
            Time duration probabilities
        """

        if X.shape[1] != 2:
            raise ValueError("Shape must be exactly (n,2) but is " + X.shape)

        x = X[:, 0]
        timestamp = X[:, 1]

        check_is_fitted(self, self.attr_to_check)

        x = column_or_1d(x)
        timestamp = column_or_1d(timestamp)

        self.time_p_ = np.ones(len(x))

        for i in range(1, len(x)):
            if x[i - 1] in self.inv_dict_labels_ and x[i] in self.inv_dict_labels_:
                time_duration = timestamp[i] - timestamp[i - 1]
                previous_event_index = self.inv_dict_labels_[x[i - 1]]
                current_event_index = self.inv_dict_labels_[x[i]]

                self.time_p_[i] = self.sda_matrix_[previous_event_index, current_event_index].get_amount(time_duration,
                                                                                                         normalized=True)

        return self.time_p_

    def predict(self, X):
        """
        For each transition in sequence X, returns a boolean value, which is True if the time
        duration probability between 2 states is more than the threshold (default: 0.5) and False if not.

        Parameters
        ----------
        X : array-like, shape = (n_samples, 2)
            First column corresponds to states, second column to timestamps

        Returns
        -------
        bool_vector : array-like
            Collection of boolean values

        """
        return self.predict_proba(X) > self.predict_threshold


class MetaSequenceMiner(BaseEstimator):
    """ Meta-model for sequence mining.
    
    This class uses MarkovSequenceMiner (transition probabilities between states) and TimeDurationSequenceMiner 
    (time duration distributions between states) within a given event window to predict states probability occurence.
        
    Attributes
    ----------
    predict_threshold : float, (default=0.5)
        Used by the predict method, to decide if the probablity outputs a True value (strictly above the threshold)
        or a False value (equal or below the threshold)
    
    markovseqminer : MarkovSequenceMiner
        Used to compute the transition probabilities between states.
        
    timeseqminer : TimeDurationSequenceMiner
        Used to compute the time duration distributions between states
    
    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.markov import MetaSequenceMiner
    
    >>> df = pd.DataFrame({'event': ['a', 'a', 'b', 'b', 'a', 'a', 'a', 'b'], \
                           'timestamp': [0, 12, 14, 29, 40, 42, 44, 48] })
    >>> metaseqminer = MetaSequenceMiner()
    >>> metaseqminer.fit(df[['event', 'timestamp']].as_matrix())
    MetaSequenceMiner(bins=10)
    
    >>> df_test = pd.DataFrame({'event': ['a', 'b', 'b', 'a', 'b', 'a'], \
                                'timestamp': [0, 2, 20, 24, 27, 38]})
    >>> metaseqminer.predict_proba(df_test[['event', 'timestamp']].as_matrix()).tolist()
    [1.0, 0.45, 0.22, 0.278, 0.22220000000000004, 0.77778]
    
    """

    def __init__(self, bins=10):
        self.bins = bins
        self.markovseqminer = MarkovSequenceMiner()
        self.timeseqminer = TimeDurationSequenceMiner(bins=self.bins)
        self.predict_threshold = 0.5

    def fit(self, X, y=None):
        """
        Fit the meta-model by estimating probabilities transition and time duration distribution between states. 

        Parameters
        ----------
        X : array-like, shape = (n_samples, 2)
            First column corresponds to states, second column to timestamps
        y : array-like, shape (n_samples,)
            No used, only here for compatibility reason

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        if X.shape[1] != 2:
            raise ValueError("Shape must be exactly (n,2) but is " + X.shape)

        self.markovseqminer.fit(X[:, 0])
        self.timeseqminer.fit(X)

        return self

    def predict_proba(self, X, operation=np.mean):
        """
        Return an array corresponding to probabilities for having these transitions and time durations
        between states of X.

        Parameters
        ----------
        X : array-like, shape = (n_samples, 2)
            First column corresponds to states, second column to timestamps
            
        operation : callable, optional (default: numpy.mean)
            Set the function used to merge transitions and time durations probabilities.
        
        Returns
        -------
        time_p_: array-like
            Mean between transition probabilities and time duration probabilities
        """

        if X.shape[1] != 2:
            raise ValueError("Shape must be exactly (n,2) but is " + X.shape)

        markov_probas, _ = self.markovseqminer.predict_proba(X[:, 0])
        sda_probas = self.timeseqminer.predict_proba(X)

        df_probas = pd.DataFrame({'markov_p': markov_probas,
                                  'sda_p': sda_probas})

        df_probas['score'] = df_probas.apply(operation, axis=1)

        return df_probas['score'].values

    def predict(self, X):
        """
        For each transition in sequence X, returns a boolean value, which is True if predict_proba result
         is more than the threshold (default: 0.5) and False if not.

        Parameters
        ----------
        X : array-like, shape = (n_samples, 2)
            First column corresponds to states, second column to timestamps

        Returns
        -------
        bool_vector : array-like
            Collection of boolean values

        """
        return self.predict_proba(X) > self.predict_threshold



import unittest

import pandas as pd
from dsbox.ml.text import LSH
from pandas.util.testing import assert_frame_equal


class TestLSH(unittest.TestCase):
    def test_get_hash(self):
        # given
        clf = LSH(n_hash_functions=3, n_bands=3)
        clf._seeds = [1, 2, 3]

        # when
        hashes = []
        for i in range(0, len(clf._seeds)):
            hashes.append(clf._get_hash(i, 'toto'))

        # then
        hashes_expected = [10143814839887274982, 8018333638267422442, 9635205340518008203]
        self.assertListEqual(hashes_expected, hashes)

    def test_minhash(self):
        # given
        clf = LSH(n_hash_functions=3, n_bands=3)
        clf._seeds = [1, 2, 3]

        # when
        # string 'octo technology' manually shinglized
        shingles = ['oct', 'cto', 'to ', 'o t', ' te', 'tec', 'ech', 'chn', 'hno', 'nol', 'olo', 'log', 'ogy']
        minhashes = list(clf._minhash(shingles))

        # then
        minhashes_expected = [3.8649734935051917e+17, 2.1739818134115853e+17, 1.4207452988806428e+18]
        self.assertListEqual(minhashes_expected, minhashes)

    def test_build_hashes(self):
        # given
        clf = LSH(n_hash_functions=3, n_bands=3)
        clf._seeds = [1, 2, 3]

        # when
        texts = ['octo', 'technology']
        hashes = clf._build_hashes(texts, 3)

        # then
        self.assertListEqual([6.6663064401848566e+18, 2.1739818134115853e+17, 4.2542837112525629e+18], list(hashes[0]))
        self.assertListEqual([3.8649734935051917e+17, 1.2233874801797798e+18, 1.4207452988806428e+18], list(hashes[1]))

    def test_reverse_hash(self):
        # given
        clf = LSH(n_hash_functions=3, n_bands=3)
        clf._seeds = [1, 2, 3]  # fixing seeds for test

        # when
        # 2 string 'octo', 'technology' manually shinglized
        minhashes = []
        shingles = ['oct', 'cto']
        minhashes.append(list(clf._minhash(shingles)))
        shingles = ['tec', 'ech', 'chn', 'hno', 'nol', 'olo', 'log', 'ogy']
        minhashes.append(list(clf._minhash(shingles)))

        reverse_hashes = clf._reverse_hash(minhashes, 3, 1)

        # then
        reverse_hashes_expected = [
            {6.6663064401848566e+18: set([0]), 3.8649734935051917e+17: set([1])},
            {2.1739818134115853e+17: set([0]), 1.2233874801797798e+18: set([1])},
            {4.2542837112525629e+18: set([0]), 1.4207452988806428e+18: set([1])}]

        self.assertListEqual(reverse_hashes_expected, reverse_hashes)

    def test_candidates(self):
        # given
        texts = ["Luc, la force en toi tu dois trouver",
                 "Trouver la force en toi tu dois Luc",
                 "Luc, Tu dois trouver la force en toi",
                 "Trouver tu dois la force en toi Luc",
                 "Tu dois la force trouver en toi",

                 "Un anneau pour les gouverner tous",
                 "Un anneau pour tous les gouverner",
                 "Tous les gouverner par le pouvoir de l'anneau",
                 "L'anneau unique pour dans les ténèbres les lier",
                 ]

        # when
        clf = LSH(n_hash_functions=99, n_bands=33)
        clf._seeds = list(range(0, 99))  # fixing seeds for test
        clf.fit(texts)

        texts_to_compare = ["La force je dois trouver en moi",
                            "L'anneau de pouvoir pour tous les gouverner"
                            ]

        candidates = clf._get_candidates(texts_to_compare)

        # then

        candidates_expected = [[4, 1, 2, 3], [7, 6, 4, 1, 2, 3, 5]]
        self.assertSetEqual(set(candidates_expected[0]), set(candidates[0]))
        self.assertSetEqual(set(candidates_expected[1]), set(candidates[1]))

        # just for displaying test results
        """for i in range(0, len(texts_to_compare)):
            print(texts_to_compare[i])
            print('--------')
            for k in candidates[i]:
                print(texts[k])

            print()"""

    def test_predict(self):
        # given
        texts = ["trouver la force en toi tu dois luc",
                 "Luc, Tu dois trouver la force en toi",
                 "Trouver tu dois la force en toi Luc",
                 "Tu dois la force trouver en toi",

                 "Un anneau pour les gouverner tous",
                 "Un anneau pour tous les gouverner",
                 "Tous les gouverner par le pouvoir de l'anneau",
                 "L'anneau unique pour dans les ténèbres les lier",
                 ]

        # when
        clf = LSH(n_hash_functions=200, n_bands=50)
        clf._seeds = list(range(0, 200))  # fixing seeds for test
        clf.fit(texts)

        texts_to_compare = ["La force je dois trouver en moi",
                            "L'anneau de pouvoir pour tous les gouverner"
                            ]

        candidates = clf.predict_proba(texts_to_compare)

        # then

        candidates_expected = pd.DataFrame([
            ['Luc, Tu dois trouver la force en toi', 0.717948717948718],
            ["Tous les gouverner par le pouvoir de l'anneau", 0.7959183673469388]])

        assert_frame_equal(candidates_expected, candidates)



import unittest

import numpy as np
from dsbox.ml.neural_networks.processing import Text2Sequence
from nltk.stem.snowball import EnglishStemmer

import logging

logging.getLogger("tensorflow").setLevel(logging.WARNING)

np.random.seed(42)


class TestText2Sequence(unittest.TestCase):
    def test_TestText2Sequence_fit_and_transform_should_return_correct_sequences(self):
        # given
        X = np.array(['this is really awesome !',
                      'this is really crap !'])

        # when
        text2seq = Text2Sequence(stemmer=EnglishStemmer())
        sequences = text2seq.fit_transform(X)

        # then
        self.assertEqual(sequences[0][0], sequences[1][0])
        self.assertEqual(sequences[0][1], sequences[1][1])
        self.assertEqual(sequences[0][2], sequences[1][2])
        self.assertNotEqual(sequences[0][3], sequences[1][3])
        self.assertEqual(sequences[0][4], sequences[1][4])


if __name__ == '__main__':
    unittest.main()



from sklearn.base import BaseEstimator, TransformerMixin
import types
import numpy as np

from sklearn.utils.validation import check_is_fitted

__author__ = "Vincent Levorato, Eric Biernat"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def project_continuous_on_categorical(dataframe, cat_col, cont_col, operation='mean'):
    return dataframe.groupby(cat_col)[cont_col].transform(operation)


class CategoricalProjector(BaseEstimator, TransformerMixin):
    """
    Aggregate continuous variables on categorical ones.

    Parameters
    ----------   
        categories_to_project: list
            List of the columns to be used as category for the projection.
            
        continous_cols: list
            List of the columns to be used as continuous variables for the projection.
        
        operation: str or function, optional, (default='mean')
            Operation used for aggregation.
            
    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.feature_engineering import CategoricalProjector
    
    >>> df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B'], \
                           'town': ['Paris', 'London', 'Paris', 'Paris', 'Roma'], \
                           'price': [0.0, 1.0, 2.0, 2.0, 1.0], \
                           'quantity': [3, 4, 1, 2, 1] \
                           })
    >>> cat_projector = CategoricalProjector(['item'], ['price', 'quantity'])
    >>> cat_projector.fit_transform(df)
      item  mean_price_per_item  mean_quantity_per_item
    0    A             0.500000                3.500000
    1    A             0.500000                3.500000
    2    B             1.666667                1.333333
    3    B             1.666667                1.333333
    4    B             1.666667                1.333333
    """

    def __init__(self, categories_to_project, continous_cols, operation='mean'):
        self.categories_to_project = categories_to_project
        self.continous_cols = continous_cols
        self.operation = operation

    def fit(self, X=None, y=None):
        """
        No-op.
        This method doesn't do anything. It exists purely for compatibility
        with the scikit-learn transformer API.

        Parameters
        ----------
            X: array-like
            y: array-like

        Returns
        -------
            self

        """

        return self

    def transform(self, X):
        """
       Transform a dataframe to build projection from continuous variables on categorical ones. Column
       names are named using the aggregation function name (i.e mean_xxx_per_xxx)

       Parameters
       ----------
           X: dataframe
               Input pandas dataframe.

       Returns
       -------
           X: dataframe 
               Dataframe with continuous on categorical projection

       """

        X_result = X[self.categories_to_project]

        operation_name = self.operation

        if isinstance(self.operation, types.FunctionType):
            operation_name = self.operation.__name__

        for cat_col in self.categories_to_project:
            for cont_col in self.continous_cols:
                X_result.loc[:, operation_name + '_{}_per_{}'.format(cont_col, cat_col)] = \
                    project_continuous_on_categorical(X, cat_col, cont_col, operation=self.operation)

        return X_result


class TagEncoder(BaseEstimator, TransformerMixin):
    """
    Alternative to sklearn LabelEncoder. The main difference is the transform operation is not crashing if
    an unknown value is encountered.

    Parameters
    ----------
        missing_value: int, optional, default: -1
            value used to encode unknown categories during the transform operation
            
    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.feature_engineering import TagEncoder

    >>> df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B']})
    >>> tagencoder = TagEncoder()
    >>> tagencoder.fit_transform(df['item'])
    [0, 0, 1, 1, 1]
    """

    def __init__(self, missing_value=-1):
        self.dict_id2tag = {}
        self.dict_tag2id = {}
        self.missing_value = missing_value

    def fit(self, X, y=None):
        """
        Fit tag encoder
        
        Parameters
        ----------
        X : array-like of shape (n_samples,)
            Target values.
            
        Returns
        -------
        self : returns an instance of self.
        """
        self.id_num_ = 0
        self.dict_id2tag = {}
        self.dict_tag2id = {}

        for x in X:
            if x not in self.dict_tag2id:
                self.dict_tag2id[x] = self.id_num_
                self.dict_id2tag[self.id_num_] = x
                self.id_num_ += 1

        if self.missing_value >= 0 and self.id_num_ >= self.missing_value:
            raise ValueError("Missing value id category is equal to one of the id category.")

        return self

    def transform(self, X):
        """
        Transform labels to normalized encoding.
        
        Parameters
        ----------
        X : array-like of shape [n_samples]
           Target values.
        Returns
        -------
        y : array-like of shape [n_samples]
        """
        check_is_fitted(self)
        x_transform = []
        for x in X:
            if x not in self.dict_tag2id:
                x_transform.append(self.missing_value)
            else:
                x_transform.append(self.dict_tag2id[x])

        return x_transform

    def inverse_transform(self, y):
        """Transform labels back to original encoding.
        Parameters
        ----------
        y : numpy array of shape [n_samples]
            Target values.
        Returns
        -------
        y : numpy array of shape [n_samples]
        """
        check_is_fitted(self)
        x_transform = []
        for x in y:
            if x not in self.dict_id2tag:
                x_transform.append(np.nan)
            else:
                x_transform.append(self.dict_id2tag[x])

        return x_transform



import logging
import string
import sys
from collections import deque

import numpy as np
from scipy.special import logsumexp
from sklearn.base import BaseEstimator
from sklearn.utils import check_array, check_random_state

import pyximport
from sklearn.utils.validation import check_is_fitted

pyximport.install(setup_args={'include_dirs': np.get_include()})

from . import _hmmc
from .utils import normalize, log_normalize, iter_from_X_lengths, log_mask_zero


_log = logging.getLogger(__name__)
#: Supported decoder algorithms.
DECODER_ALGORITHMS = frozenset(("viterbi", "map"))


class ConvergenceMonitor:
    """Monitors and reports convergence to :data:`sys.stderr`.

    Parameters
    ----------
    tol : double
        Convergence threshold. EM has converged either if the maximum
        number of iterations is reached or the log probability
        improvement between the two consecutive iterations is less
        than threshold.

    n_iter : int
        Maximum number of iterations to perform.

    verbose : bool
        If ``True`` then per-iteration convergence reports are printed,
        otherwise the monitor is mute.

    Attributes
    ----------
    history : deque
        The log probability of the data for the last two training
        iterations. If the values are not strictly increasing, the
        model did not converge.

    iter : int
        Number of iterations performed while training the model.

    Examples
    --------
    Use custom convergence criteria by subclassing ``ConvergenceMonitor``
    and redefining the ``converged`` method. The resulting subclass can
    be used by creating an instance and pointing a model's ``monitor_``
    attribute to it prior to fitting.

    >>> from dsbox.ml.markov.hmmlearn.base import ConvergenceMonitor
    >>> from dsbox.ml.markov.hmmlearn import hmm
    >>>
    >>> class ThresholdMonitor(ConvergenceMonitor):
    ...     @property
    ...     def converged(self):
    ...         return (self.iter == self.n_iter or
    ...                 self.history[-1] >= self.tol)
    >>>
    >>> model = hmm.GaussianHMM(n_components=2, tol=5, verbose=True)
    >>> model.monitor_ = ThresholdMonitor(model.monitor_.tol,
    ...                                   model.monitor_.n_iter,
    ...                                   model.monitor_.verbose)
    """
    _template = "{iter:>10d} {logprob:>16.4f} {delta:>+16.4f}"

    def __init__(self, tol, n_iter, verbose):
        self.tol = tol
        self.n_iter = n_iter
        self.verbose = verbose
        self.history = deque(maxlen=2)
        self.iter = 0

    def __repr__(self):
        class_name = self.__class__.__name__
        params = sorted(dict(vars(self), history=list(self.history)).items())
        return ("{}(\n".format(class_name)
                + "".join(map("    {}={},\n".format, *zip(*params)))
                + ")")

    def _reset(self):
        """Reset the monitor's state."""
        self.iter = 0
        self.history.clear()

    def report(self, logprob):
        """Reports convergence to :data:`sys.stderr`.

        The output consists of three columns: iteration number, log
        probability of the data at the current iteration and convergence
        rate.  At the first iteration convergence rate is unknown and
        is thus denoted by NaN.

        Parameters
        ----------
        logprob : float
            The log probability of the data as computed by EM algorithm
            in the current iteration.
        """
        if self.verbose:
            delta = logprob - self.history[-1] if self.history else np.nan
            message = self._template.format(
                iter=self.iter + 1, logprob=logprob, delta=delta)
            print(message, file=sys.stderr)

        self.history.append(logprob)
        self.iter += 1

    @property
    def converged(self):
        """``True`` if the EM algorithm converged and ``False`` otherwise."""
        # XXX we might want to check that ``logprob`` is non-decreasing.
        return (self.iter == self.n_iter or
                (len(self.history) == 2 and
                 self.history[1] - self.history[0] < self.tol))


class _BaseHMM(BaseEstimator):
    r"""Base class for Hidden Markov Models.

    This class allows for easy evaluation of, sampling from, and
    maximum a posteriori estimation of the parameters of a HMM.

    See the instance documentation for details specific to a
    particular object.

    Parameters
    ----------
    n_components : int
        Number of states in the model.

    startprob_prior : array, shape (n_components, ), optional
        Parameters of the Dirichlet prior distribution for
        :attr:`startprob_`.

    transmat_prior : array, shape (n_components, n_components), optional
        Parameters of the Dirichlet prior distribution for each row
        of the transition probabilities :attr:`transmat_`.

    algorithm : string, optional
        Decoder algorithm. Must be one of "viterbi" or "map".
        Defaults to "viterbi".

    random_state: RandomState or an int seed, optional
        A random number generator instance.

    n_iter : int, optional
        Maximum number of iterations to perform.

    tol : float, optional
        Convergence threshold. EM will stop if the gain in log-likelihood
        is below this value.

    verbose : bool, optional
        When ``True`` per-iteration convergence reports are printed
        to :data:`sys.stderr`. You can diagnose convergence via the
        :attr:`monitor_` attribute.

    params : string, optional
        Controls which parameters are updated in the training
        process.  Can contain any combination of 's' for startprob,
        't' for transmat, and other characters for subclass-specific
        emission parameters. Defaults to all parameters.

    init_params : string, optional
        Controls which parameters are initialized prior to
        training.  Can contain any combination of 's' for
        startprob, 't' for transmat, and other characters for
        subclass-specific emission parameters. Defaults to all
        parameters.

    Attributes
    ----------
    monitor\_ : ConvergenceMonitor
        Monitor object used to check the convergence of EM.

    startprob\_ : array, shape (n_components, )
        Initial state occupation distribution.

    transmat\_ : array, shape (n_components, n_components)
        Matrix of transition probabilities between states.
    """
    def __init__(self, n_components=1,
                 startprob_prior=1.0, transmat_prior=1.0,
                 algorithm="viterbi", random_state=None,
                 n_iter=10, tol=1e-2, verbose=False,
                 params=string.ascii_letters,
                 init_params=string.ascii_letters):
        self.n_components = n_components
        self.params = params
        self.init_params = init_params
        self.startprob_prior = startprob_prior
        self.transmat_prior = transmat_prior
        self.algorithm = algorithm
        self.random_state = random_state
        self.n_iter = n_iter
        self.tol = tol
        self.verbose = verbose
        self.monitor_ = ConvergenceMonitor(self.tol, self.n_iter, self.verbose)

    def get_stationary_distribution(self):
        """Compute the stationary distribution of states.
        """
        # The stationary distribution is proportional to the left-eigenvector
        # associated with the largest eigenvalue (i.e., 1) of the transition
        # matrix.
        check_is_fitted(self, "transmat_")
        eigvals, eigvecs = np.linalg.eig(self.transmat_.T)
        eigvec = np.real_if_close(eigvecs[:, np.argmax(eigvals)])
        return eigvec / eigvec.sum()

    def score_samples(self, X, lengths=None):
        """Compute the log probability under the model and compute posteriors.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, ), optional
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.

        Returns
        -------
        logprob : float
            Log likelihood of ``X``.

        posteriors : array, shape (n_samples, n_components)
            State-membership probabilities for each sample in ``X``.

        See Also
        --------
        score : Compute the log probability under the model.
        decode : Find most likely state sequence corresponding to ``X``.
        """
        check_is_fitted(self, "startprob_")
        self._check()

        X = check_array(X)
        n_samples = X.shape[0]
        logprob = 0
        posteriors = np.zeros((n_samples, self.n_components))
        for i, j in iter_from_X_lengths(X, lengths):
            framelogprob = self._compute_log_likelihood(X[i:j])
            logprobij, fwdlattice = self._do_forward_pass(framelogprob)
            logprob += logprobij

            bwdlattice = self._do_backward_pass(framelogprob)
            posteriors[i:j] = self._compute_posteriors(fwdlattice, bwdlattice)
        return logprob, posteriors

    def score(self, X, lengths=None):
        """Compute the log probability under the model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, ), optional
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.

        Returns
        -------
        logprob : float
            Log likelihood of ``X``.

        See Also
        --------
        score_samples : Compute the log probability under the model and
            posteriors.
        decode : Find most likely state sequence corresponding to ``X``.
        """
        check_is_fitted(self, "startprob_")
        self._check()

        X = check_array(X)
        # XXX we can unroll forward pass for speed and memory efficiency.
        logprob = 0
        for i, j in iter_from_X_lengths(X, lengths):
            framelogprob = self._compute_log_likelihood(X[i:j])
            logprobij, _fwdlattice = self._do_forward_pass(framelogprob)
            logprob += logprobij
        return logprob

    def _decode_viterbi(self, X):
        framelogprob = self._compute_log_likelihood(X)
        return self._do_viterbi_pass(framelogprob)

    def _decode_map(self, X):
        _, posteriors = self.score_samples(X)
        logprob = np.max(posteriors, axis=1).sum()
        state_sequence = np.argmax(posteriors, axis=1)
        return logprob, state_sequence

    def decode(self, X, lengths=None, algorithm=None):
        """Find most likely state sequence corresponding to ``X``.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, ), optional
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.

        algorithm : string
            Decoder algorithm. Must be one of "viterbi" or "map".
            If not given, :attr:`decoder` is used.

        Returns
        -------
        logprob : float
            Log probability of the produced state sequence.

        state_sequence : array, shape (n_samples, )
            Labels for each sample from ``X`` obtained via a given
            decoder ``algorithm``.

        See Also
        --------
        score_samples : Compute the log probability under the model and
            posteriors.
        score : Compute the log probability under the model.
        """
        check_is_fitted(self, "startprob_")
        self._check()

        algorithm = algorithm or self.algorithm
        if algorithm not in DECODER_ALGORITHMS:
            raise ValueError("Unknown decoder {!r}".format(algorithm))

        decoder = {
            "viterbi": self._decode_viterbi,
            "map": self._decode_map
        }[algorithm]

        X = check_array(X)
        n_samples = X.shape[0]
        logprob = 0
        state_sequence = np.empty(n_samples, dtype=int)
        for i, j in iter_from_X_lengths(X, lengths):
            # XXX decoder works on a single sample at a time!
            logprobij, state_sequenceij = decoder(X[i:j])
            logprob += logprobij
            state_sequence[i:j] = state_sequenceij

        return logprob, state_sequence

    def predict(self, X, lengths=None):
        """Find most likely state sequence corresponding to ``X``.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, ), optional
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.

        Returns
        -------
        state_sequence : array, shape (n_samples, )
            Labels for each sample from ``X``.
        """
        _, state_sequence = self.decode(X, lengths)
        return state_sequence

    def predict_proba(self, X, lengths=None):
        """Compute the posterior probability for each state in the model.

        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, ), optional
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.

        Returns
        -------
        posteriors : array, shape (n_samples, n_components)
            State-membership probabilities for each sample from ``X``.
        """
        _, posteriors = self.score_samples(X, lengths)
        return posteriors

    def sample(self, n_samples=1, random_state=None):
        """Generate random samples from the model.

        Parameters
        ----------
        n_samples : int
            Number of samples to generate.

        random_state : RandomState or an int seed
            A random number generator instance. If ``None``, the object's
            ``random_state`` is used.

        Returns
        -------
        X : array, shape (n_samples, n_features)
            Feature matrix.

        state_sequence : array, shape (n_samples, )
            State sequence produced by the model.
        """
        check_is_fitted(self, "startprob_")
        self._check()

        if random_state is None:
            random_state = self.random_state
        random_state = check_random_state(random_state)

        startprob_cdf = np.cumsum(self.startprob_)
        transmat_cdf = np.cumsum(self.transmat_, axis=1)

        currstate = (startprob_cdf > random_state.rand()).argmax()
        state_sequence = [currstate]
        X = [self._generate_sample_from_state(
            currstate, random_state=random_state)]

        for t in range(n_samples - 1):
            currstate = (transmat_cdf[currstate] > random_state.rand()) \
                .argmax()
            state_sequence.append(currstate)
            X.append(self._generate_sample_from_state(
                currstate, random_state=random_state))

        return np.atleast_2d(X), np.array(state_sequence, dtype=int)

    def fit(self, X, lengths=None):
        """Estimate model parameters.

        An initialization step is performed before entering the
        EM algorithm. If you want to avoid this step for a subset of
        the parameters, pass proper ``init_params`` keyword argument
        to estimator's constructor.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, )
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.

        Returns
        -------
        self : object
            Returns self.
        """
        X = check_array(X)
        self._init(X, lengths=lengths)
        self._check()

        self.monitor_._reset()
        for iter in range(self.n_iter):
            stats = self._initialize_sufficient_statistics()
            curr_logprob = 0
            for i, j in iter_from_X_lengths(X, lengths):
                framelogprob = self._compute_log_likelihood(X[i:j])
                logprob, fwdlattice = self._do_forward_pass(framelogprob)
                curr_logprob += logprob
                bwdlattice = self._do_backward_pass(framelogprob)
                posteriors = self._compute_posteriors(fwdlattice, bwdlattice)
                self._accumulate_sufficient_statistics(
                    stats, X[i:j], framelogprob, posteriors, fwdlattice,
                    bwdlattice)

            # XXX must be before convergence check, because otherwise
            #     there won't be any updates for the case ``n_iter=1``.
            self._do_mstep(stats)

            self.monitor_.report(curr_logprob)
            if self.monitor_.converged:
                break

        if (self.transmat_.sum(axis=1) == 0).any():
            _log.warning("Some rows of transmat_ have zero sum because no "
                         "transition from the state was ever observed.")

        return self

    def _do_viterbi_pass(self, framelogprob):
        n_samples, n_components = framelogprob.shape
        state_sequence, logprob = _hmmc._viterbi(
            n_samples, n_components, log_mask_zero(self.startprob_),
            log_mask_zero(self.transmat_), framelogprob)
        return logprob, state_sequence

    def _do_forward_pass(self, framelogprob):
        n_samples, n_components = framelogprob.shape
        fwdlattice = np.zeros((n_samples, n_components))
        _hmmc._forward(n_samples, n_components,
                       log_mask_zero(self.startprob_),
                       log_mask_zero(self.transmat_),
                       framelogprob, fwdlattice)
        with np.errstate(under="ignore"):
            return logsumexp(fwdlattice[-1]), fwdlattice

    def _do_backward_pass(self, framelogprob):
        n_samples, n_components = framelogprob.shape
        bwdlattice = np.zeros((n_samples, n_components))
        _hmmc._backward(n_samples, n_components,
                        log_mask_zero(self.startprob_),
                        log_mask_zero(self.transmat_),
                        framelogprob, bwdlattice)
        return bwdlattice

    def _compute_posteriors(self, fwdlattice, bwdlattice):
        # gamma is guaranteed to be correctly normalized by logprob at
        # all frames, unless we do approximate inference using pruning.
        # So, we will normalize each frame explicitly in case we
        # pruned too aggressively.
        log_gamma = fwdlattice + bwdlattice
        log_normalize(log_gamma, axis=1)
        with np.errstate(under="ignore"):
            return np.exp(log_gamma)

    def _needs_init(self, code, name):
        if code in self.init_params:
            if hasattr(self, name):
                _log.warning(
                    "Even though the %r attribute is set, it will be "
                    "overwritten during initialization because 'init_params' "
                    "contains %r", name, code)
            return True
        if not hasattr(self, name):
            return True
        return False

    def _get_n_fit_scalars_per_param(self):
        """Return a mapping of fittable parameter name (as in ``self.params``)
        to the number of corresponding scalar parameters that will actually be
        fitted.

        This is used to detect whether the user did not pass enough data points
        for a non-degenerate fit.
        """

    def _init(self, X, lengths):
        """Initializes model parameters prior to fitting.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        lengths : array-like of integers, shape (n_sequences, )
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.
        """
        init = 1. / self.n_components
        if self._needs_init("s", "startprob_"):
            self.startprob_ = np.full(self.n_components, init)
        if self._needs_init("t", "transmat_"):
            self.transmat_ = np.full((self.n_components, self.n_components),
                                     init)
        n_fit_scalars_per_param = self._get_n_fit_scalars_per_param()
        if n_fit_scalars_per_param is not None:
            n_fit_scalars = sum(
                n_fit_scalars_per_param[p] for p in self.params)
            if X.size < n_fit_scalars:
                _log.warning(
                    "Fitting a model with %d free scalar parameters with only "
                    "%d data points will result in a degenerate solution.",
                    n_fit_scalars, X.size)

    def _check(self):
        """Validates model parameters prior to fitting.

        Raises
        ------

        ValueError
            If any of the parameters are invalid, e.g. if :attr:`startprob_`
            don't sum to 1.
        """
        self.startprob_ = np.asarray(self.startprob_)
        if len(self.startprob_) != self.n_components:
            raise ValueError("startprob_ must have length n_components")
        if not np.allclose(self.startprob_.sum(), 1.0):
            raise ValueError("startprob_ must sum to 1.0 (got {:.4f})"
                             .format(self.startprob_.sum()))

        self.transmat_ = np.asarray(self.transmat_)
        if self.transmat_.shape != (self.n_components, self.n_components):
            raise ValueError(
                "transmat_ must have shape (n_components, n_components)")
        if not np.allclose(self.transmat_.sum(axis=1), 1.0):
            raise ValueError("rows of transmat_ must sum to 1.0 (got {})"
                             .format(self.transmat_.sum(axis=1)))

    def _compute_log_likelihood(self, X):
        """Computes per-component log probability under the model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.

        Returns
        -------
        logprob : array, shape (n_samples, n_components)
            Log probability of each sample in ``X`` for each of the
            model states.
        """

    def _generate_sample_from_state(self, state, random_state=None):
        """Generates a random sample from a given component.

        Parameters
        ----------
        state : int
            Index of the component to condition on.

        random_state: RandomState or an int seed
            A random number generator instance. If ``None``, the object's
            ``random_state`` is used.

        Returns
        -------
        X : array, shape (n_features, )
            A random sample from the emission distribution corresponding
            to a given component.
        """

    # Methods used by self.fit()

    def _initialize_sufficient_statistics(self):
        """Initializes sufficient statistics required for M-step.

        The method is *pure*, meaning that it doesn't change the state of
        the instance.  For extensibility computed statistics are stored
        in a dictionary.

        Returns
        -------
        nobs : int
            Number of samples in the data.

        start : array, shape (n_components, )
            An array where the i-th element corresponds to the posterior
            probability of the first sample being generated by the i-th
            state.

        trans : array, shape (n_components, n_components)
            An array where the (i, j)-th element corresponds to the
            posterior probability of transitioning between the i-th to j-th
            states.
        """
        stats = {'nobs': 0,
                 'start': np.zeros(self.n_components),
                 'trans': np.zeros((self.n_components, self.n_components))}
        return stats

    def _accumulate_sufficient_statistics(self, stats, X, framelogprob,
                                          posteriors, fwdlattice, bwdlattice):
        """Updates sufficient statistics from a given sample.

        Parameters
        ----------
        stats : dict
            Sufficient statistics as returned by
            :meth:`~base._BaseHMM._initialize_sufficient_statistics`.

        X : array, shape (n_samples, n_features)
            Sample sequence.

        framelogprob : array, shape (n_samples, n_components)
            Log-probabilities of each sample under each of the model states.

        posteriors : array, shape (n_samples, n_components)
            Posterior probabilities of each sample being generated by each
            of the model states.

        fwdlattice, bwdlattice : array, shape (n_samples, n_components)
            Log-forward and log-backward probabilities.
        """
        stats['nobs'] += 1
        if 's' in self.params:
            stats['start'] += posteriors[0]
        if 't' in self.params:
            n_samples, n_components = framelogprob.shape
            # when the sample is of length 1, it contains no transitions
            # so there is no reason to update our trans. matrix estimate
            if n_samples <= 1:
                return

            log_xi_sum = np.full((n_components, n_components), -np.inf)
            _hmmc._compute_log_xi_sum(n_samples, n_components, fwdlattice,
                                      log_mask_zero(self.transmat_),
                                      bwdlattice, framelogprob,
                                      log_xi_sum)
            with np.errstate(under="ignore"):
                stats['trans'] += np.exp(log_xi_sum)

    def _do_mstep(self, stats):
        """Performs the M-step of EM algorithm.

        Parameters
        ----------
        stats : dict
            Sufficient statistics updated from all available samples.
        """
        # If a prior is < 1, `prior - 1 + starts['start']` can be negative.  In
        # that case maximization of (n1+e1) log p1 + ... + (ns+es) log ps under
        # the conditions sum(p) = 1 and all(p >= 0) show that the negative
        # terms can just be set to zero.
        # The ``np.where`` calls guard against updating forbidden states
        # or transitions in e.g. a left-right HMM.
        if 's' in self.params:
            startprob_ = np.maximum(self.startprob_prior - 1 + stats['start'],
                                    0)
            self.startprob_ = np.where(self.startprob_ == 0, 0, startprob_)
            normalize(self.startprob_)
        if 't' in self.params:
            transmat_ = np.maximum(self.transmat_prior - 1 + stats['trans'], 0)
            self.transmat_ = np.where(self.transmat_ == 0, 0, transmat_)
            normalize(self.transmat_, axis=1)


import unittest

import numpy as np
from dsbox.ml.explain.prediction_interval import compute_prediction_interval
from sklearn.datasets import load_boston
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split

# set random seed for test
np.random.seed(42)


class PredictionIntervalTest(unittest.TestCase):
    def test_90_percentile_prediction_interval_has_90_percent_of_ground_truth_inside_the_interval(self):
        # Given
        boston = load_boston()
        data_train, data_eval, target_train, target_eval = train_test_split(
            boston['data'][:100], boston['target'][:100], test_size=0.3, random_state=42)
        rf = RandomForestRegressor(n_estimators=80, n_jobs=-1, random_state=42)
        rf.fit(data_train, target_train)

        # When
        lower_error_list, upper_error_list = compute_prediction_interval(rf, data_eval, percentile=90)

        # Then
        ground_truth_inside_interval = 0
        for i, target_value in enumerate(target_eval):
            if lower_error_list[i] <= target_value <= upper_error_list[i]:
                ground_truth_inside_interval += 1
        ground_truth_inside_interval_percentage = ground_truth_inside_interval / float(len(target_eval))
        # check that we have roughly 90% of ground truth value inside the computed interval
        self.assertAlmostEqual(ground_truth_inside_interval_percentage, 0.90, 2)

    def test_compute_prediction_interval_with_stub_model(self):
        # Given
        class TreeStub:
            def predict(self, data_row):
                return [1]

        class RfStub:
            estimators_ = [TreeStub() for _ in range(100)]

        data_eval = np.array([[1, 2, 3], [1, 2, 4]])
        expected_lower_error_list = [1, 1]
        expected_upper_error_list = [1, 1]

        # When
        lower_error_list, upper_error_list = compute_prediction_interval(RfStub(), data_eval, percentile=90)

        # Then
        self.assertEqual(lower_error_list, expected_lower_error_list)
        self.assertEqual(upper_error_list, expected_upper_error_list)



# Copyright 2013 Lars Buitinck / University of Amsterdam
# encoding: utf-8

import numpy as np
from scipy.sparse import csr_matrix


def make_trans_matrix(y, n_classes, dtype=np.float64):
    """Make a sparse transition matrix for y.

    Takes a label sequence y and returns an indicator matrix with n_classes²
    columns of the label transitions in y: M[i, j, k] means y[i-1] == j and
    y[i] == k. The first row will be empty.
    """
    indices = np.empty(len(y), dtype=np.int32)

    for i in range(len(y) - 1):
        indices[i] = y[i] * i + y[i + 1]

    indptr = np.arange(len(y) + 1)
    indptr[-1] = indptr[-2]

    return csr_matrix((np.ones(len(y), dtype=dtype), indices, indptr),
                      shape=(len(y), n_classes ** 2))



import unittest

import pandas as pd
from pandas.util.testing import assert_series_equal, assert_frame_equal

from dsbox.ml.feature_engineering import project_continuous_on_categorical, CategoricalProjector, TagEncoder


class TestCategorical(unittest.TestCase):
    def test_project_continuous_on_categorical_function_should_return_correct_results(self):
        # given
        df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B'],
                           'price': [0.0, 1.0, 2.0, 2.0, 1.0]})

        # when
        serie_transformed = project_continuous_on_categorical(df, 'item', 'price')

        # then
        serie_expected = pd.Series(data=[0.5, 0.5, 1.66666667, 1.66666667, 1.66666667],
                                   index=df.index,
                                   name='price')

        assert_series_equal(serie_expected, serie_transformed)

    def test_categoricalprojector_should_project_continuous_on_categorical_data_properly(self):
        # given
        df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B'],
                           'town': ['Paris', 'London', 'Paris', 'Paris', 'Roma'],
                           'price': [0.0, 1.0, 2.0, 2.0, 1.0],
                           'quantity': [3, 4, 1, 2, 1]
                           })

        # when
        cat_projector = CategoricalProjector(['item', 'town'], ['price', 'quantity'])
        df_transformed = cat_projector.fit_transform(df)

        # then
        df_expected = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B'],
                                    'town': ['Paris', 'London', 'Paris', 'Paris', 'Roma'],
                                    'mean_price_per_item': [0.5, 0.5, 1.666667, 1.666667, 1.666667],
                                    'mean_quantity_per_item': [3.5, 3.5, 1.3333333333333333, 1.3333333333333333,
                                                               1.3333333333333333],
                                    'mean_price_per_town': [1.3333333333333333, 1.0, 1.3333333333333333,
                                                            1.3333333333333333, 1.0],
                                    'mean_quantity_per_town': [2, 4, 2, 2, 1]
                                    })

        df_transformed = df_transformed[df_expected.columns]

        assert_frame_equal(df_expected, df_transformed)

    def test_tagencoder_should_encode_categories_nicely(self):
        # given
        df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B']})

        # when
        tagencoder = TagEncoder()
        column_encoded = tagencoder.fit_transform(df['item'])

        # then
        column_expected = [0, 0, 1, 1, 1]

        self.assertListEqual(column_expected, column_encoded)

    def test_tagencoder_should_encode_unknown_categories_without_crashing_like_this_shitty_sklearn_labelencoder(self):
        # given
        df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B']})
        df_bis = pd.DataFrame({'item': ['A', 'A', 'B', 'C', 'B', 'D']})

        # when
        tagencoder = TagEncoder(missing_value=-1)
        tagencoder.fit(df['item'])
        column_encoded = tagencoder.transform(df_bis['item'])

        # then
        column_expected = [0, 0, 1, -1, 1, -1]

        self.assertListEqual(column_expected, column_encoded)

    def test_tagencoder_return_decoded_labels(self):
        # given
        df = pd.DataFrame({'item': ['A', 'A', 'B', 'B', 'B']})

        # when
        tagencoder = TagEncoder()
        column_encoded = tagencoder.fit_transform(df['item'])
        column_decoded = tagencoder.inverse_transform(column_encoded)

        # then
        column_expected = ['A', 'A', 'B', 'B', 'B']

        self.assertListEqual(column_expected, column_decoded)


if __name__ == '__main__':
    unittest.main()



import numpy as np

__credits__ = "https://github.com/tmadl/sklearn-random-rotation-ensembles"


def random_rot_mat_incl_flip(n):
    r = np.random.normal(size=(n, n))
    Q, R = np.linalg.qr(r)  # A=QR
    M = np.dot(Q, np.diag(np.sign(np.diag(R))))  # diag(R)>0
    return M.astype(np.float32)


def random_rotation_matrix(n):
    M = random_rot_mat_incl_flip(n)
    if np.linalg.det(M) < 0:
        M[:, 0] = -M[:, 0]  # det(M)=1
    return M



import unittest

import numpy as np
import pandas as pd

from dsbox.ml.markov import MarkovSequenceMiner, TimeDurationSequenceMiner, MetaSequenceMiner


class TestMarkovSequenceMiner(unittest.TestCase):
    def test_chained_probas_with_transition_proba_matrix_prefilled(self):
        # given
        pmatrix = np.array([[0.5, 0.25, 0.25],
                             [0.5, 0, 0.5],
                             [0.25, 0.25, 0.5]])

        # when
        clf = MarkovSequenceMiner(equiproba=True)
        clf.transit_matrix_ = pmatrix

        clf.labels_ = np.array(['Rain', 'Nice', 'Snow'])
        clf.inv_dict_labels_ = {clf.labels_[i]: i for i in range(0, len(clf.labels_))}
        clf.init_vector = np.ones(len(clf.labels_)) * (1.0 / len(clf.labels_))

        clf.predict_proba(['Nice', 'Nice', 'Rain', 'Rain'])
        proba_vector = clf.cur_vector.tolist()[0]

        # then
        values_to_obtain = np.array([0.406, 0.188, 0.406])

        self.assertEqual(np.round(proba_vector[0], 3), values_to_obtain[0])
        self.assertEqual(np.round(proba_vector[1], 3), values_to_obtain[1])
        self.assertEqual(np.round(proba_vector[2], 3), values_to_obtain[2])

    def test_fit_predict_proba_simple_markov_chain_should_return_correct_values(self):
        # given
        sequence = ['a', 'a', 'b', 'c', 'd', 'a']

        # when
        clf = MarkovSequenceMiner()
        clf.fit(sequence)
        clf.predict_proba(['a', 'b', 'c', 'd', 'e', 'a'])
        y = clf.cur_vector.tolist()[0]

        # then
        y_true = [0.531, 0.281, 0.062, 0.125]
        self.assertEqual(np.round(y[0], 3), y_true[0])
        self.assertEqual(np.round(y[1], 3), y_true[1])
        self.assertEqual(np.round(y[2], 3), y_true[2])

    def test_fit_predict_proba_simple_markov_chain_with_warm_start_ie_keep_fitted_state(self):
        # given
        sequence = ['a', 'a', 'b', 'c', 'd', 'a']

        # when
        clf = MarkovSequenceMiner(warm_start=True)
        clf.fit(sequence)
        clf.predict_proba(['a', 'b', 'c', 'd', 'e'])
        clf.predict_proba(['e', 'a'])
        y = clf.cur_vector.tolist()[0]

        # then
        y_true = [0.531, 0.281, 0.062, 0.125]
        self.assertEqual(np.round(y[0], 3), y_true[0])
        self.assertEqual(np.round(y[1], 3), y_true[1])
        self.assertEqual(np.round(y[2], 3), y_true[2])

    def test_chained_transition_probas_which_are_computed_during_fitting(self):
        # given
        sequence = ['a', 'a', 'b', 'c', 'd', 'a']

        # when
        clf = MarkovSequenceMiner()
        clf.fit(sequence)
        chained_p, transition_p = clf.predict_proba(['a', 'b', 'c', 'd', 'e', 'a'])

        # then
        chained_p_true = [1.0, 0.5, 0.5, 0.5, 0.0, 0.0]
        transition_p_true = [1.0, 0.5, 1.0, 1.0, 0.0, 0.0]

        self.assertListEqual(chained_p.tolist(), chained_p_true)
        self.assertListEqual(transition_p.tolist(), transition_p_true)

    def test_fit_predict_which_should_return_correct_anomalies(self):
        # given
        sequence = ['a', 'a', 'b', 'c', 'a', 'd', 'd', 'd', 'a']

        # when
        clf = MarkovSequenceMiner()
        clf.fit(sequence)
        y = clf.predict(['c', 'b', 'd', 'd', 'a']).tolist()

        # then
        y_true = [False, False, True, True, True]
        self.assertEqual(y, y_true)

    def test_fit_predict_next_which_should_return_next_most_likely(self):
        # given
        sequence = ['a', 'a', 'b', 'c', 'a', 'd', 'd', 'd', 'a']

        # when
        clf = MarkovSequenceMiner()
        clf.fit(sequence)
        y = clf.predict_next(['c', 'b', 'd', 'd', 'a'])

        # then
        y_true = ['a']
        self.assertEqual(y, y_true)


class TestTimeDurationSequenceMiner(unittest.TestCase):
    def test_fit_predict_probas_should_return_correct_values(self):
        # given
        df = pd.DataFrame({'event': ['a', 'a', 'b', 'b', 'a', 'a', 'a', 'b'],
                           'timestamp': [0, 12, 14, 29, 40, 42, 44, 48]
                           })

        # when
        tds = TimeDurationSequenceMiner()
        tds.fit(df[['event', 'timestamp']].values)

        df_test = pd.DataFrame({'event': ['a', 'b', 'b', 'a', 'b', 'a'],
                                'timestamp': [0, 2, 20, 24, 27, 38]
                                })

        y_test = tds.predict_proba(df_test[['event', 'timestamp']].values)

        # then
        y_true = [1, 0.5, 0, 0, 0, 1]
        self.assertListEqual(y_true, y_test.tolist())

    def test_fit_predict_should_return_correct_anomalies(self):
        # given
        df = pd.DataFrame({'event': ['a', 'a', 'b', 'b', 'a', 'a', 'a', 'b'],
                           'timestamp': [0, 12, 14, 29, 40, 42, 44, 48]
                           })

        # when
        tds = TimeDurationSequenceMiner()
        tds.fit(df[['event', 'timestamp']].values)

        df_test = pd.DataFrame({'event': ['a', 'b', 'b', 'a', 'b', 'a'],
                                'timestamp': [0, 2, 20, 24, 27, 38]
                                })

        y_test = tds.predict(df_test[['event', 'timestamp']].values)

        # then
        y_true = [True, False, False, False, False, True]
        self.assertListEqual(y_true, y_test.tolist())


class TestMetaSequenceMiner(unittest.TestCase):
    def test_metasequenceminer_fit_predict_proba_should_return_correct_values(self):
        # given
        df = pd.DataFrame({'event': ['a', 'a', 'b', 'b', 'a', 'a', 'a', 'b'],
                           'timestamp': [0, 12, 14, 29, 40, 42, 44, 48]
                           })

        # when
        metaseqminer = MetaSequenceMiner()
        metaseqminer.fit(df[['event', 'timestamp']].values)

        df_test = pd.DataFrame({'event': ['a', 'b', 'b', 'a', 'b', 'a'],
                                'timestamp': [0, 2, 20, 24, 27, 38]
                                })

        y_test = metaseqminer.predict_proba(df_test[['event', 'timestamp']].values)

        # then
        y_true = [1.0, 0.45, 0.22, 0.28, 0.22, 0.78]
        self.assertListEqual(y_true, y_test.round(2).tolist())



__author__ = "Vincent Levorato"
__license__ = "Apache 2.0"


def check_estimator_predict_proba(estimator):
    if not hasattr(estimator, "predict_proba"):
        raise RuntimeError('classifier has no predict_proba method.')


from collections import Counter

import numpy as np
import pandas as pd
from dsbox.ml.outliers import fft_outliers, mad_outliers
from sklearn.base import BaseEstimator, ClusterMixin, OutlierMixin
from sklearn.cluster import DBSCAN
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.covariance import EmpiricalCovariance, MinCovDet
from sklearn.exceptions import NotFittedError
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.mixture import BayesianGaussianMixture
from sklearn.mixture._base import BaseMixture
from sklearn.neighbors._kde import KernelDensity
from sklearn.utils.validation import check_is_fitted, column_or_1d

from scipy.stats import norm, median_absolute_deviation

__author__ = "Vincent Levorato, Rémy Frenoy"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class CovarianceOutliers(BaseEstimator, OutlierMixin):
    """ Covariance wrapper outlier estimator.
    
    Encapsulates sklearn covariance estimator to extract outliers from Mahalanobis squared distance.
    
    Parameters
    ----------
    
    cov_estimator : EmpiricalCovariance, optional (default=MinCovDet())
    
    threshold : float (default=None)
        Used by predict method : if probability returned by predict_proba method is above this value, the element 
        is considered as an outlier. If not set, it takes the mean of probabilities.
    
    
    Attributes
    ----------
    
    mahal_dist_ : array
       Mahalanobis squared distances computed by the covariance estimator. 
       
       
    Examples
    --------
    >>> import pandas as pd
    >>> from sklearn.covariance import MinCovDet
    >>> from dsbox.ml.outliers import CovarianceOutliers
    
    >>> X = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])
    >>> cov_outliers = CovarianceOutliers(MinCovDet())
    >>> outliers = cov_outliers.fit_predict(X)
    >>> outliers
    array([False, False, False, False, False, False,  True,  True, False,
           False, False, False, False, False])
    
    """

    def __init__(self, cov_estimator=MinCovDet(), threshold=None):
        if not isinstance(cov_estimator, EmpiricalCovariance):
            raise TypeError("Estimator must be a sklearn.covariance.EmpiricalCovariance class")

        self.cov_estimator = cov_estimator
        self.threshold = threshold

        self.attr_to_check = ["mahal_dist_"]

    def fit(self, X, y=None):
        """
        Fits the covariance model according to the given training data and parameters.
        
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : not used, present for API consistence purpose.

        Returns
        -------
        self : object
            Returns self.
        """
        self.cov_estimator.fit(X)
        self.mahal_dist_ = self.cov_estimator.mahalanobis(X)

        return self

    def predict_proba(self, X=None):
        """
        Returns probability for each element to be an outlier. 
        Exactly, it computes the ratio of the Mahalanobis squared distance relative to the distance sum.
        
        Parameters
        ----------
        X : not used, present for API consistence purpose.

        Returns
        -------
        Probabilities for each element to be an outlier

        """
        check_is_fitted(self, self.attr_to_check)

        return self.mahal_dist_ / np.sum(self.mahal_dist_)

    def predict(self, X=None):
        """
        Returns a boolean tag for each element to be an outlier. It takes each Mahalanobis squared distance,
        transforms it in probability, and checks if it exceeds the threshold attribute.
        
        Parameters
        ----------
        X : not used, present for API consistence purpose.

        Returns
        -------
        Boolean array with outlier tag
        """
        probas = self.predict_proba(X)

        if self.threshold is None:
            self.threshold = np.mean(probas)
        return probas > self.threshold


class GaussianProcessOutliers(BaseEstimator, OutlierMixin):
    """ Gaussian Process wrapper outlier estimator

    Encapsulates sklearn Gaussian Process Regressor estimator to extract outliers which should be
    out of the confidence intervals along the "infinite" normal marginal distributions estimated by the
    GP estimator.

    Parameters
    ----------

    gp_estimator : GaussianProcessRegressor, optional (default=GaussianProcessRegressor(alpha=0.9, normalize_y=True))

    n_samples : int, optional (default=500)
        Corresponds to the amount of points used for the prediction (should be less than the original amount of elements)
        
    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from sklearn.gaussian_process import GaussianProcessRegressor
    >>> from dsbox.ml.outliers import GaussianProcessOutliers

    >>> np.random.seed(42)
    >>> data = np.random.random_sample(200) * 2 - 1
    >>> data[150] = 5
    >>> data[190] = -6
    >>> df = pd.DataFrame(data)
    
    >>> gp_outliers = GaussianProcessOutliers(GaussianProcessRegressor(alpha=0.95, normalize_y=True), n_samples=100)
    >>> _ = gp_outliers.fit(df)
    >>> outliers = gp_outliers.predict(df, confidence=0.999)
    >>> outlier_positions = np.argwhere(outliers == np.amax(outliers)).flatten().tolist()
    >>> outlier_positions
    [150, 190]
    
    """

    def __init__(self, gp_estimator=GaussianProcessRegressor(alpha=0.9, normalize_y=True), n_samples=500):

        if not isinstance(gp_estimator, GaussianProcessRegressor):
            raise TypeError("Estimator must be a sklearn.gaussian_process.GaussianProcessRegressor class")

        self.gp_estimator = gp_estimator
        self.n_samples = n_samples

    def fit(self, X, y=None):
        """
        Fits the Gaussian Process estimator according to the given training data and parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : not used, present for API consistence purpose.

        Returns
        -------
        self : object
        """

        y = column_or_1d(X)
        x = np.arange(0, y.size)
        self.gp_estimator.fit(x.reshape(-1, 1), y)

        return self

    def predict(self, X, confidence=0.9999):
        """
        Returns a boolean tag for each element to be an outlier. It uses the predict method of the 
        Gaussian Process estimator, smoothing the inputs value, and comparing the prediction according to
        confidence interval.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
        
        confidence : float, optional (default=0.9999)

        Returns
        -------
        Boolean array with outlier tag
        """
        y = column_or_1d(X)
        x = np.arange(0, y.size)

        if self.gp_estimator.normalize_y:
            y_origin_mean = 0
            y_origin_std = 1
        else:
            y_origin_mean = np.mean(y)
            y_origin_std = np.std(y)

        X_pred = pd.DataFrame(np.linspace(x.min(), x.max(), num=self.n_samples))
        y_mean, y_std = self.gp_estimator.predict(X_pred, return_std=True)
        tolerance = norm.ppf(confidence, y_origin_mean, y_origin_std)
        indices = [int(x) for x in np.linspace(0, 99, y.size)]

        lower_bound = y_mean[indices] - tolerance * y_std[indices]
        upper_bound = y_mean[indices] + tolerance * y_std[indices]

        outliers = (y < lower_bound) | (y > upper_bound)
        return outliers


class KMeansOneClusterOutliers(BaseEstimator, OutlierMixin):
    """ KMeans One Cluster Outliers estimator
    
    This estimator uses distance information to categorize outliers. In this case, KMeans algorithm is used,
    computing only one cluster and one centroid. Depending on the distance to the centroid, elements are categorized
    as outliers or not (the greater the distance of an element, the more likely it is to be considered an outlier).
    
    Parameters
    ----------
    
    kmeans_estimator : KMeans, optional (default=KMeans(n_clusters=1))
    
    threshold : float (default=None)
        Used by predict method : if probability returned by predict_proba method is above this value, the element 
        is considered as an outlier. If not set, it takes the mean of probabilities.
    
    
    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.outliers import KMeansOneClusterOutliers
    
    >>> df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])
    >>> kmoc_outliers = KMeansOneClusterOutliers()
    >>> outliers = kmoc_outliers.fit_predict(df)
    >>> outliers
    array([False, False, False, False, False, False,  True,  True, False,
           False, False, False, False, False])
    
    """

    def __init__(self, kmeans_estimator=KMeans(n_clusters=1), threshold=None):
        if not isinstance(kmeans_estimator, KMeans) and not isinstance(kmeans_estimator, MiniBatchKMeans):
            raise TypeError("Estimator must be a sklearn.cluster.KMeans or MiniBatchKMeans class")

        if kmeans_estimator.get_params()['n_clusters'] != 1:
            raise ValueError("KMeans estimator n_clusters parameter must be set to 1")

        self.kmeans_estimator = kmeans_estimator
        self.threshold = threshold

    def fit(self, X, y=None):
        """
        Fits the kmeans one cluster model according to the given training data and parameters.
        
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
           Training data, where n_samples is the number of samples
           and n_features is the number of features.
        y : not used, present for API consistence purpose.
        
        Returns
        -------
        self : object
           Returns self.
        """
        self.kmeans_estimator.fit(X)
        return self

    def predict_proba(self, X):
        """
        Returns probability for each element to be an outlier. 
        Exactly, it computes the ratio of the distance to the unique centroid with the distance sum.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Probabilities for each element to be an outlier

        """

        distances = self.kmeans_estimator.transform(X).ravel()
        sum_distances = np.sum(distances)
        if sum_distances == 0:
            return distances
        return distances / float(sum_distances)

    def predict(self, X):
        """
        Returns a boolean tag for each element to be an outlier. It takes each distance to the unique centroid,
        transform it in probability, and checks if it exceeds the threshold attribute.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Boolean array with outlier tag
        """
        probas = self.predict_proba(X)
        if self.threshold is None:
            self.threshold = np.mean(probas)
        return probas > self.threshold


class KMeansIterativeOneClusterOutliers(BaseEstimator, OutlierMixin):
    """ KMeans Iterative One Cluster Outliers estimator

    This estimator uses iteratively KMeans One Cluster Outliers estimator to find outliers. It applies n times
    the KMOC estimator, finding and removing outliers from the original dataset, giving them a tag which correspond
    to the iteration number they have been removed.

    Parameters
    ----------

    kmoc_estimator : KMeansOneClusterOutliers, optional (default=KMeansOneClusterOutliers())

    n_iterations : int, optional, (default=5)
        Max iterations amount used for finding outliers with KMOC estimator iteratively.

    Examples
    --------
    >>> import pandas as pd
    >>> from dsbox.ml.outliers import KMeansIterativeOneClusterOutliers
    
    >>> df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])
    >>> kmoc_outliers = KMeansIterativeOneClusterOutliers()
    >>> outliers = kmoc_outliers.fit_predict(df)
    >>> outliers
    array([ 2.,  1.,  1.,  2.,  2., -1.,  0.,  0.,  1.,  1., -1.,  1.,  1.,
            2.])

    """

    def __init__(self, kmoc_estimator=KMeansOneClusterOutliers(), n_iterations=5):
        if not isinstance(kmoc_estimator, KMeansOneClusterOutliers):
            raise TypeError("Estimator must be a dsbox.ml.outliers.KMeansOneClusterOutliers class")

        self.kmoc_estimator = kmoc_estimator
        self.n_iterations = n_iterations

    def fit_predict(self, X, y=None):
        """
        Fits the kmeans one cluster model according to the given training data and parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
           Training data, where n_samples is the number of samples
           and n_features is the number of features.
        y : not used, present for API consistence purpose.

        Returns
        -------
        self : object
           Returns self.
        """
        XX = X.copy()
        outlier_iter = np.ones(len(XX)) * -1
        i = 0

        while i < self.n_iterations and len(XX) > 0:
            self.kmoc_estimator.fit(XX)
            outliers = self.kmoc_estimator.predict(XX)
            if True in outliers:
                XX['outliers'] = outliers
                outlier_positions = XX[XX['outliers'] == True].index.values
                for position in outlier_positions:
                    outlier_iter[position] = i
                del XX['outliers']
                XX = XX.drop(outlier_positions)
                i += 1
            else:
                i = self.n_iterations

        return outlier_iter


class GMMOutliers(BaseEstimator, OutlierMixin):
    """ Gaussian mixture model Outliers estimator
    
    This estimator uses density information to categorize outliers. In this case, GMM algorithm is used,
    computing an amount of components. Depending on the PDF (probability density function), elements are categorized
    as outliers or not (the less density, the more likely it is to be considered an outlier).
    
    Parameters
    ----------
    
    gmm_estimator : BaseMixture, optional (default=BayesianGaussianMixture(n_components=1))
    
    threshold : float, optional(default=None)
        Used by predict method : if probability returned by predict_proba method is above this value, the element 
        is considered as an outlier. If not set, it takes the mean of probabilities.
        
    Examples
    --------
    
    >>> import pandas as pd
    >>> from dsbox.ml.outliers import GMMOutliers
    
    >>> df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])
    >>> gmm_outliers = GMMOutliers()
    >>> outliers = gmm_outliers.fit_predict(df)
    >>> outliers
    array([False, False, False, False, False, False,  True,  True, False,
           False, False, False, False, False])
    
    """

    def __init__(self, gmm_estimator=BayesianGaussianMixture(n_components=1), threshold=None):
        if not isinstance(gmm_estimator, BaseMixture):
            raise TypeError("Estimator must be a sklearn.mixture.base.BaseMixture class")

        self.gmm_estimator = gmm_estimator
        self.threshold = threshold

    def fit(self, X, y=None):
        """
        Fits the GMM model according to the given training data and parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
           Training data, where n_samples is the number of samples
           and n_features is the number of features.
        y : not used, present for API consistence purpose.

        Returns
        -------
        self : object
           Returns self.
        """
        self.gmm_estimator.fit(X)
        return self

    def predict_proba(self, X):
        """
        Returns probability for each element to be an outlier. 
        It transforms probability densities obtained (score samples) by integrating them to 1 to probabilities
        of having this value. In this case, it returns probability for each element to be an outlier.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Probabilities for each element to be an outlier

        """

        dx = X.diff().abs().median().sum()
        return 1 - (np.exp(self.gmm_estimator.score_samples(X)) * np.power(dx, X.shape[1]))

    def predict(self, X):
        """
        Returns a boolean tag for each element to be an outlier. It takes probability density transformed
        into probability, and checks if it exceeds the threshold attribute.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Boolean array with outlier tag
        """
        probas = self.predict_proba(X)
        if self.threshold is None:
            self.threshold = np.mean(probas)
        return probas > self.threshold


class ClusteringOutliers(BaseEstimator, OutlierMixin):
    """ Clustering model Outliers estimator

    This estimator uses clustering information to categorize outliers. Depending on the strategy, this
    estimator can use un-clustering information (like DBSCAN which classify elements that not belongs to any
    cluster), frontier, size or density of the clusters.

    Parameters
    ----------

    cluster_estimator : BaseMixture or ClusterMixin, optional (default=DBSCAN())

    threshold : float, optional (default=None)
        Used by predict method : if probability returned by predict_proba method is above this value, the element 
        is considered as an outlier. If not set, it takes the mean of probabilities.
    
    strategy : {'unclustered', 'frontier', 'size', 'density'} (default='size')
        Stragegy used to identify outliers.
        
        'unclustered': if clustering estimator is able to put a label (-1) on elements out of
        all clusters (DBSCAN, MeanShift), this strategy will identify these elements as outliers.
        
        'frontier': this strategy considers, for each cluster, the farest elements of the center as outliers.
        
        'size': this strategy considers the elements of the smallest cluster(s) as outliers.
        
        'density': this strategy considers the elements of the less dense cluster(s) as outliers.
        
    kernel_density : KernelDensity
        Only used by 'density' strategy, used to compute density per cluster.
    
    Attributes
    ----------
    
    is_fitted : boolean
        Used only for consistency between clustering estimators methods.
    
    Examples
    --------
    
    >>> import pandas as pd 
    >>> from dsbox.ml.outliers import ClusteringOutliers
    
    >>> df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])
    >>> clustering_outliers = ClusteringOutliers(cluster_estimator=KMeans(n_clusters=2), strategy='size')
    >>> outliers = clustering_outliers.fit_predict(df)
    >>> outliers
    array([False, False, False, False, False, False,  True,  True, False,
           False, False, False, False, False])
     
    
    """

    def __init__(self, cluster_estimator=DBSCAN(), threshold=None, strategy='size',
                 kernel_density=KernelDensity(kernel='epanechnikov', bandwidth=0.2)):
        if not isinstance(cluster_estimator, BaseMixture) and not isinstance(cluster_estimator, ClusterMixin):
            raise TypeError("Estimator must be a sklearn.mixture.base.BaseMixture class" +
                            " or a sklearn.base.ClusterMixin class.")

        self.cluster_estimator = cluster_estimator
        self.threshold = threshold
        self.strategy = strategy
        self.kernel_density = kernel_density
        self.is_fitted = False

    def fit(self, X, y=None):
        """
        Fits the clustering model according to the given training data and parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
           Training data, where n_samples is the number of samples
           and n_features is the number of features.
        y : not used, present for API consistence purpose.

        Returns
        -------
        self : object
           Returns self.
        """
        self.cluster_estimator.fit(X)
        self.is_fitted = True
        return self

    def predict_proba(self, X):
        """
        Returns probability for each element to be an outlier, according to the chosen strategy.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Probabilities for each element to be an outlier

        """
        if not self.is_fitted:
            raise NotFittedError("This instance is not fitted yet.")

        labels = []
        if 'predict' in dir(self.cluster_estimator):
            labels = self.cluster_estimator.predict(X)
        else:
            labels = self.cluster_estimator.labels_

        probas = np.zeros(len(labels))

        if self.strategy == 'unclustered':
            return (labels == -1).astype('float')

        if self.strategy == 'frontier':
            distances = self.cluster_estimator.transform(X)
            probas_all = np.zeros(distances.shape)
            for col_id in range(0, distances.shape[1]):
                probas_all[:, col_id] = distances[:, col_id] / np.sum(distances[:, col_id])

            for i in range(0, len(labels)):
                probas[i] = probas_all[i, labels[i]]

        if self.strategy == 'size':
            X_size = float(len(X))
            counter = Counter(labels)
            size_ratio = np.zeros(len(counter.items()))
            for label in counter.keys():
                size_ratio[label] = counter[label] / X_size

            for i in range(0, len(labels)):
                probas[i] = 1 - size_ratio[labels[i]]

        if self.strategy == 'density':
            X['label_'] = labels
            score_label = {}
            for i in range(0, np.max(labels) + 1):
                self.kernel_density.fit(X[X['label_'] == i])
                score_label[i] = self.kernel_density.score(X[X['label_'] == i])

            del X['label_']

            # force the density for unclustered element to zero (if any)
            score_label[-1] = 0.0

            for i in range(0, len(labels)):
                probas[i] = 1 - (score_label[labels[i]] / sum(score_label.values()))

        return probas

    def predict(self, X):
        """
        Returns a boolean tag for each element to be an outlier. It takes predict_proba method, 
        and checks if it exceeds the threshold attribute.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Boolean array with outlier tag
        """
        probas = self.predict_proba(X)

        if self.strategy == 'unclustered':
            return probas == 1

        if self.threshold is None:
            self.threshold = np.mean(probas)
        return probas > self.threshold


class MADOutliers(BaseEstimator, OutlierMixin):
    """  Median Absolute Deviation outliers estimator
    
    A simple scikit wrappering method used in dsbox.ml.outliers.utils.mad_outliers. See documentation for
    further purpose.
    
    
    Parameters
    ----------

    window: int, optional (default=None)
        window size for MAD calculation, if None, window is equal to length of dataset
    
    cutoff : int, optional (default=2)
        amount of times residuals relative to the median exceed the ratio to the MAD
    
    
    Examples
    --------
    >>> import pandas as pd 
    >>> from dsbox.ml.outliers import MADOutliers
    
    >>> df = pd.DataFrame({'values': [1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1]})
    >>> mad_outliers = MADOutliers()
    >>> outliers = mad_outliers.fit_predict(df)
    >>> outliers
    array([False, False, False, False,  True, False,  True,  True,  True,
           False, False, False,  True, False])
    
    
    """

    def __init__(self, window=None, cutoff=2, threshold=None):
        self.window = window
        self.cutoff = cutoff
        self.threshold = threshold

    def fit(self, X, y=None):
        """
        Fits the MAD model according to the given training data and parameters.
        
        Parameters
        ----------
        X : pandas DataFrame
        y : not used, present for API consistence purpose.
        
        Returns
        -------
        self : object
          Returns self.
        """
        if self.window is None:
            self.window = len(X)
        self.X_ = pd.DataFrame()
        for column in X.columns:
            self.X_['mad_{}'.format(column)] = X[column].rolling(self.window, min_periods=1).apply(
                median_absolute_deviation, raw=False)
            self.X_['median_{}'.format(column)] = X[column].rolling(self.window, min_periods=1).median()
            self.X_[column] = (np.abs(X[column] - self.X_['median_{}'.format(column)]) / self.X_[
                'mad_{}'.format(column)]) > self.cutoff
            self.X_ = self.X_.drop(['mad_{}'.format(column), 'median_{}'.format(column)], axis=1)

        return self

    def predict_proba(self, X=None):
        """
        Returns probability for each element to be an outlier.

        Parameters
        ----------
        X : not used, present for API consistence purpose.

        Returns
        -------
        Probabilities for each element to be an outlier

        """

        return self.X_.mean(axis=1).values

    def predict(self, X):
        """
        Returns a boolean tag for each element to be an outlier. It takes predict_proba method,
        and checks if it exceeds the threshold attribute.

        Parameters
        ----------
        X : not used, present for API consistence purpose.

        Returns
        -------
        Boolean array with outlier tag
        """

        probas = self.predict_proba(X)

        if self.threshold is None:
            self.threshold = np.mean(probas)
        return probas > self.threshold


class FFTOutliers(BaseEstimator, OutlierMixin):
    """  Fast Fourier Transformation outliers estimator

    A simple scikit wrapper using bdacore.outliers.utils.fft_outliers method. See documentation for
    further purpose.


    Parameters
    ----------

    freq_cut_index : int, optional (default=None)
        set the index in the FFT numpy array beyond which the frequency is considered to be filtered. The highest,
        the less filtering. By default, it sets this index to 0.9 of x length
        
    outlier_proportion : float, optional (default=0.1)
        set the proportion of outliers to return


    Examples
    --------
    >>> import pandas as pd 
    >>> from dsbox.ml.outliers import FFTOutliers

    >>> df = pd.DataFrame([1, 0, 0, 1, 10, 2, 115, 110, 32, 16, 2, 0, 15, 1])
    >>> fft_outliers = FFTOutliers()
    >>> outliers = fft_outliers.fit_predict(df)
    >>> outliers.values
    array([False, False, False, False, False, False, False,  True, False,
           False, False, False, False, False])


    """

    def __init__(self, freq_cut_index=None, outlier_proportion=0.1, threshold=None):
        self.freq_cut_index = freq_cut_index
        self.outlier_proportion = outlier_proportion
        self.threshold = threshold

    def fit(self, X=None, y=None):
        """
        Fits the FFT model according to the given training data and parameters.
 
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
           Training data, where n_samples is the number of samples
           and n_features is the number of features.
        y : not used, present for API consistence purpose.
 
        Returns
        -------
        self : object
           Returns self.
        """

        self.X_ = pd.DataFrame(columns=X.columns)
        for column in X.columns:
            self.X_[column] = fft_outliers(X[column], freq_cut_index=self.freq_cut_index,
                                           outlier_proportion=self.outlier_proportion)

        return self

    def predict_proba(self, X=None):
        """
        Returns probability for each element to be an outlier.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Probabilities for each element to be an outlier

        """

        return self.X_.mean(axis=1)

    def predict(self, X):
        """
        Returns a boolean tag for each element to be an outlier. It takes predict_proba method, 
        and checks if it exceeds the threshold attribute.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        Returns
        -------
        Boolean array with outlier tag
        """
        probas = self.predict_proba(X)

        if self.threshold is None:
            self.threshold = np.mean(probas)
        return probas > self.threshold



import unittest

from dsbox.ml.markov.histrogramdata import HistogramData


class TestHistogramData(unittest.TestCase):
    def test_get_amount_method_should_return_correct_values(self):
        # given
        data = [55, 28, 37]
        intervals = [0, 5, 10, 15]  # histogram intervals

        # when
        hda = HistogramData(intervals, data)

        # then
        amount = hda.get_amount(6)
        self.assertEqual(amount, 28)

        amount = hda.get_amount(11)
        self.assertEqual(amount, 37)

        amount = hda.get_amount(1)
        self.assertEqual(amount, 55)

        amount = hda.get_amount(-1)
        self.assertEqual(amount, 0)

    def test_get_amount_normalized_should_return_correct_values(self):
        # given
        data = [55, 28, 37]
        intervals = [0, 5, 10, 15]  # histogram intervals

        # when
        hda = HistogramData(intervals, data)
        amount = hda.get_amount(6, normalized=True)

        # then
        self.assertEqual(amount, float(28) / sum(data))



# Copyright Lars Buitinck 2013.

"""Decoding (inference) algorithms."""

import numpy as np
import pyximport
pyximport.install(setup_args={'include_dirs': np.get_include()})

from .bestfirst import bestfirst
from .viterbi import viterbi

DECODERS = {"bestfirst": bestfirst,
            "viterbi": viterbi}



import unittest

import numpy as np


import logging

from dsbox.ml.neural_networks.keras_factory.text_models import LSTMFactory
from dsbox.ml.neural_networks.processing.workflow import TextNeuralNetPipeline, ImageNeuralNetPipeline

logging.getLogger("tensorflow").setLevel(logging.WARNING)

np.random.seed(42)


class TestPipeline(unittest.TestCase):
    def test_fit_predict_text_nn_pipeline_should_return_some_result(self):
        # given
        x_train = np.array(['this is really really awesome !',
                            'it is so awesome !',
                            'that sucks']
                           )
        y_train = np.array([1, 1, 0])

        # when
        model = TextNeuralNetPipeline(factory_class=LSTMFactory, num_labels=2)
        model.fit(x_train, y_train, verbose=0)

        x_test = np.array(['it is really awesome !'])
        y_pred = model.predict(x_test)

        # then
        self.assertIsNotNone(y_pred)

    def test_fit_predict_proba_text_nn_pipeline_should_return_some_result(self):
        # given
        x_train = np.array(['this is really really awesome !',
                            'it is so awesome !',
                            'that sucks']
                           )
        y_train = np.array([1, 1, 0])

        # when
        model = TextNeuralNetPipeline(factory_class=LSTMFactory, num_labels=2)
        model.fit(x_train, y_train, verbose=0)

        x_test = np.array(['it is really awesome !'])
        y_pred = model.predict_proba(x_test)[0]

        # then
        self.assertIsNotNone(y_pred)

    def test_fit_image_nn_workflow_should_set_params_automatically(self):
        # given
        workflow = ImageNeuralNetPipeline(weights="imagenet")

        # when
        workflow.fit()

        # then
        self.assertTupleEqual((299, 299), workflow.img_size_)
        self.assertEqual("block14_sepconv2_act", workflow.last_conv_layer_name_)
        self.assertListEqual(["avg_pool", "predictions"], workflow.classifier_layer_names_)




import importlib

from dsbox.ml.neural_networks import KerasFactory

__author__ = "Vincent Levorato"
__license__ = "Apache 2.0"


class KerasApplicationFactory(KerasFactory):
    def create_model(self, module_name='tensorflow.keras.applications.xception', model_name='Xception', **kwargs):
        """
        Build a keras pre-trained model


        Parameters
        ----------
        module_name: str, optional (default='tensorflow.keras.applications')
            Keras module containing pre-built models
        model_name: str, optional (default='Xception')
            Keras pre-built model
        **kwargs :
            parameters passed to inner keras model

        Returns
        -------
        Keras model

        .. Keras pre-trained models list: https://keras.io/api/applications/

        """
        module = importlib.import_module(module_name)
        model_builder = getattr(module, model_name)

        return model_builder(**kwargs)
